from __future__ import annotations

import math
from pathlib import Path
import random
from typing import Protocol

import pyray as rl
from raylib import defines as rd

from grim.audio import AudioState, update_audio
from grim.assets import PaqTextureCache, load_paq_entries
from grim.config import CrimsonConfig
from grim.fonts.grim_mono import GrimMonoFont, draw_grim_mono_text, load_grim_mono_font
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width
from grim.geom import Vec2
from grim.math import clamp

from grim.rand import Crand
from .creatures.spawn import RANDOM_HEADING_SENTINEL
from .game_world import GameWorld
from .gameplay import PlayerInput, PlayerState, weapon_assign_player
from .ui.cursor import draw_menu_cursor
from .ui.perk_menu import UiButtonState, UiButtonTextureSet, button_draw, button_update, button_width
from .weapons import WEAPON_TABLE

WORLD_SIZE = 1024.0
DEMO_VARIANT_COUNT = 6

_DEMO_UPSELL_MESSAGES: tuple[str, ...] = (
    "Want more Levels?",
    "Want more Weapons?",
    "Want more Perks?",
    "Want unlimited Play time?",
    "Want to post your high scores?",
)

DEMO_PURCHASE_URL = "http://buy.crimsonland.com"
DEMO_PURCHASE_SCREEN_LIMIT_MS = 16_000
DEMO_PURCHASE_INTERSTITIAL_LIMIT_MS = 10_000

_DEMO_PURCHASE_TITLE = "Upgrade to the full version of Crimsonland Today!"
_DEMO_PURCHASE_FEATURES_TITLE = "Full version features:"
_DEMO_PURCHASE_FEATURE_LINES: tuple[tuple[str, float], ...] = (
    ("-Unlimited Play Time in three thrilling Game Modes!", 22.0),
    ("-The varied weapon arsenal consisting of over 20 unique", 17.0),
    (" weapons that allow you to deal death with plasma, lead,", 17.0),
    (" fire and electricity!", 22.0),
    ("-Over 40 game altering Perks!", 22.0),
    ("-40 insane Levels that give you", 18.0),
    (" hours of intense and fun gameplay!", 22.0),
    ("-The ability to post your high scores online!", 44.0),
)
_DEMO_PURCHASE_FOOTER = "Purchasing the game is very easy and secure."


class DemoState(Protocol):
    assets_dir: Path
    rng: random.Random
    config: CrimsonConfig
    texture_cache: PaqTextureCache | None
    audio: AudioState | None
    preserve_bugs: bool
    quit_requested: bool


def _weapon_name(weapon_id: int) -> str:
    for weapon in WEAPON_TABLE:
        if weapon.weapon_id == weapon_id:
            return weapon.name or f"weapon_{weapon_id}"
    return f"weapon_{weapon_id}"


class DemoView:
    """Attract-mode demo scaffold.

    Modeled after the classic demo helpers in crimsonland.exe:
      - demo_setup_variant_0 @ 0x00402ED0
      - demo_setup_variant_1 @ 0x004030F0
      - demo_setup_variant_2 @ 0x00402FE0
      - demo_setup_variant_3 @ 0x00403250
      - demo_mode_start       @ 0x00403390
    """

    def __init__(self, state: DemoState) -> None:
        self._state = state
        self._world = GameWorld(
            assets_dir=state.assets_dir,
            world_size=WORLD_SIZE,
            demo_mode_active=True,
            hardcore=bool(int(state.config.data.get("hardcore_flag", 0) or 0)),
            difficulty_level=0,
            preserve_bugs=bool(state.preserve_bugs),
            texture_cache=state.texture_cache,
            config=state.config,
            audio=state.audio,
            audio_rng=state.rng,
        )
        self._crand = Crand(0)
        self._demo_targets: list[int | None] = []
        self._variant_index = 0
        self._demo_variant_index = 0
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._finished = False
        self._upsell_message_index = 0
        self._upsell_pulse_ms = 0
        self._upsell_font: GrimMonoFont | None = None
        self._small_font: SmallFontData | None = None
        self._purchase_active = False
        self._purchase_url_opened = False
        self._purchase_button = UiButtonState("Purchase", force_wide=True)
        self._maybe_later_button = UiButtonState("Maybe later", force_wide=True)
        self._spawn_rng = Crand(0)

    def open(self) -> None:
        self._finished = False
        self._upsell_message_index = 0
        self._upsell_pulse_ms = 0
        self._purchase_active = False
        self._purchase_url_opened = False
        self._purchase_button = UiButtonState("Purchase", force_wide=True)
        self._maybe_later_button = UiButtonState("Maybe later", force_wide=True)
        self._variant_index = 0
        self._demo_variant_index = 0
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._crand.srand(self._state.rng.getrandbits(32))
        self._world.open()
        self._demo_mode_start()

    def close(self) -> None:
        self._world.close()
        if self._upsell_font is not None:
            rl.unload_texture(self._upsell_font.texture)
            self._upsell_font = None
        if self._small_font is not None:
            rl.unload_texture(self._small_font.texture)
            self._small_font = None

    def is_finished(self) -> bool:
        return self._finished

    def update(self, dt: float) -> None:
        if self._state.audio is not None:
            update_audio(self._state.audio, dt)
        if self._finished:
            return
        frame_dt = min(dt, 0.1)
        frame_dt_ms = int(frame_dt * 1000.0)
        if frame_dt_ms <= 0:
            return

        if (
            (not self._purchase_active)
            and getattr(self._state, "demo_enabled", False)
            and self._purchase_screen_triggered()
        ):
            self._begin_purchase_screen(DEMO_PURCHASE_SCREEN_LIMIT_MS, reset_timeline=False)

        if self._purchase_active:
            self._upsell_pulse_ms += frame_dt_ms
            self._update_purchase_screen(frame_dt_ms)
            self._quest_spawn_timeline_ms += frame_dt_ms
            if self._quest_spawn_timeline_ms > self._demo_time_limit_ms:
                # demo_purchase_screen_update restarts the demo once the purchase screen
                # timer exceeds demo_time_limit_ms.
                self._demo_mode_start()
            return

        if self._skip_triggered():
            self._finished = True
            return

        self._quest_spawn_timeline_ms += frame_dt_ms
        self._update_world(frame_dt)
        if self._quest_spawn_timeline_ms > self._demo_time_limit_ms:
            self._demo_mode_start()

    def draw(self) -> None:
        if self._purchase_active:
            self._draw_purchase_screen()
            return
        self._world.draw()
        self._draw_overlay()

    def _skip_triggered(self) -> bool:
        if rl.get_key_pressed() != 0:
            return True
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            return True
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_RIGHT):
            return True
        return False

    def _purchase_screen_triggered(self) -> bool:
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            return True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            return True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_SPACE):
            return True
        return False

    def _begin_purchase_screen(self, limit_ms: int, *, reset_timeline: bool) -> None:
        self._purchase_active = True
        if reset_timeline:
            self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = max(0, int(limit_ms))
        self._purchase_url_opened = False
        self._purchase_button = UiButtonState("Purchase", force_wide=True)
        self._maybe_later_button = UiButtonState("Maybe later", force_wide=True)

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _purchase_var_28_2(self) -> float:
        screen_w = int(self._state.config.screen_width)
        if screen_w == 0x320:  # 800
            return 64.0
        if screen_w == 0x400:  # 1024
            return 128.0
        return 0.0

    def _update_purchase_screen(self, dt_ms: int) -> None:
        dt_ms = max(0, int(dt_ms))
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._purchase_active = False
            self._finished = True
            return

        font = self._ensure_small_font()
        cache = self._ensure_cache()
        textures = UiButtonTextureSet(
            button_sm=cache.get_or_load("ui_buttonSm", "ui/ui_button_64x32.jaz").texture,
            button_md=cache.get_or_load("ui_buttonMd", "ui/ui_button_128x32.jaz").texture,
        )
        if textures.button_sm is None and textures.button_md is None:
            return

        w = float(self._state.config.screen_width)
        h = float(self._state.config.screen_height)
        wide_shift = self._purchase_var_28_2()
        button_base_y = h / 2.0 + 102.0 + wide_shift * 0.3
        button_base_pos = Vec2(w / 2.0 + 128.0, button_base_y + 50.0)

        mouse = rl.get_mouse_position()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        scale = 1.0
        button_w = button_width(
            font, self._purchase_button.label, scale=scale, force_wide=self._purchase_button.force_wide
        )
        if button_update(
            self._purchase_button,
            pos=button_base_pos,
            width=float(button_w),
            dt_ms=float(dt_ms),
            mouse=mouse,
            click=bool(click),
        ):
            if not self._purchase_url_opened:
                self._purchase_url_opened = True
                try:
                    import webbrowser

                    webbrowser.open(DEMO_PURCHASE_URL)
                except Exception:
                    pass
            if hasattr(self._state, "quit_requested"):
                self._state.quit_requested = True

        if button_update(
            self._maybe_later_button,
            pos=button_base_pos.offset(dy=40.0),
            width=float(button_w),
            dt_ms=float(dt_ms),
            mouse=mouse,
            click=bool(click),
        ):
            self._purchase_active = False
            self._finished = True
            return

        # Keyboard activation for convenience; original uses UI mouse.
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER):
            if not self._purchase_url_opened:
                self._purchase_url_opened = True
                try:
                    import webbrowser

                    webbrowser.open(DEMO_PURCHASE_URL)
                except Exception:
                    pass
            if hasattr(self._state, "quit_requested"):
                self._state.quit_requested = True

        # Keep referenced to avoid unused warnings if this method grows.
        _ = textures

    def _draw_purchase_screen(self) -> None:
        rl.clear_background(rl.BLACK)

        logos = getattr(self._state, "logos", None)
        if logos is None or logos.backplasma.texture is None:
            return
        backplasma = logos.backplasma.texture

        pulse_phase = float(self._upsell_pulse_ms % 1000)
        pulse = math.sin(pulse_phase * 6.2831855)
        pulse = pulse * pulse

        screen_w = float(self._state.config.screen_width)
        screen_h = float(self._state.config.screen_height)

        # demo_purchase_screen_update @ 0x0040b985:
        #   - full-screen quad
        #   - UV: 0..0.5 (top-left quarter of the backplasma atlas)
        #   - per-corner color slots, with a sin^2 pulse at bottom-right

        def _to_u8(value: float) -> int:
            return int(clamp(value, 0.0, 1.0) * 255.0 + 0.5)

        c0 = rl.Color(_to_u8(0.0), _to_u8(0.0), _to_u8(0.0), _to_u8(1.0))
        c1 = rl.Color(_to_u8(0.0), _to_u8(0.0), _to_u8(0.3), _to_u8(1.0))
        c2 = rl.Color(
            _to_u8(0.0),
            _to_u8(0.4),
            _to_u8(pulse * 0.55),
            _to_u8(pulse),
        )
        c3 = rl.Color(_to_u8(0.0), _to_u8(0.4), _to_u8(0.4), _to_u8(1.0))

        rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)
        rl.rl_set_texture(backplasma.id)
        rl.rl_begin(rd.RL_QUADS)
        # TL
        rl.rl_color4ub(c0.r, c0.g, c0.b, c0.a)
        rl.rl_tex_coord2f(0.0, 0.0)
        rl.rl_vertex2f(0.0, 0.0)
        # TR
        rl.rl_color4ub(c1.r, c1.g, c1.b, c1.a)
        rl.rl_tex_coord2f(0.5, 0.0)
        rl.rl_vertex2f(screen_w, 0.0)
        # BR
        rl.rl_color4ub(c2.r, c2.g, c2.b, c2.a)
        rl.rl_tex_coord2f(0.5, 0.5)
        rl.rl_vertex2f(screen_w, screen_h)
        # BL
        rl.rl_color4ub(c3.r, c3.g, c3.b, c3.a)
        rl.rl_tex_coord2f(0.0, 0.5)
        rl.rl_vertex2f(0.0, screen_h)
        rl.rl_end()
        rl.rl_set_texture(0)
        rl.end_blend_mode()

        wide_shift = self._purchase_var_28_2()

        # Mockup and logo textures.
        if logos.mockup.texture is not None:
            mockup = logos.mockup.texture
            x = screen_w / 2.0 - 128.0 + wide_shift
            y = screen_h / 2.0 - 140.0
            dst = rl.Rectangle(x, y, 512.0, 256.0)
            src = rl.Rectangle(0.0, 0.0, float(mockup.width), float(mockup.height))
            rl.draw_texture_pro(mockup, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

        if logos.cl_logo.texture is not None:
            cl_logo = logos.cl_logo.texture
            x = screen_w / 2.0 - 256.0
            y = screen_h / 2.0 - 200.0 - wide_shift * 0.4
            dst = rl.Rectangle(x, y, 512.0, 64.0)
            src = rl.Rectangle(0.0, 0.0, float(cl_logo.width), float(cl_logo.height))
            rl.draw_texture_pro(cl_logo, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

        small = self._ensure_small_font()
        text_scale = 1.2
        x_text = screen_w / 2.0 - 296.0 - wide_shift * 0.8
        y = screen_h / 2.0 - 104.0
        color = rl.Color(255, 255, 255, 255)
        draw_small_text(small, _DEMO_PURCHASE_TITLE, Vec2(x_text, y), text_scale, color)
        y += 28.0
        draw_small_text(small, _DEMO_PURCHASE_FEATURES_TITLE, Vec2(x_text, y), text_scale, color)

        underline_w = measure_small_text_width(small, _DEMO_PURCHASE_FEATURES_TITLE, text_scale)
        rl.draw_rectangle_rec(rl.Rectangle(x_text, y + 15.0, underline_w, 2.0), rl.Color(255, 255, 255, 160))

        y += 22.0
        x_list = x_text + 8.0
        for line, delta_y in _DEMO_PURCHASE_FEATURE_LINES:
            draw_small_text(small, line, Vec2(x_list, y), text_scale, color)
            y += delta_y
        draw_small_text(small, _DEMO_PURCHASE_FOOTER, Vec2(x_text, y), text_scale, color)

        # Buttons on the right.
        cache = self._ensure_cache()
        textures = UiButtonTextureSet(
            button_sm=cache.get_or_load("ui_buttonSm", "ui/ui_button_64x32.jaz").texture,
            button_md=cache.get_or_load("ui_buttonMd", "ui/ui_button_128x32.jaz").texture,
        )
        if textures.button_sm is None and textures.button_md is None:
            return

        button_base_y = screen_h / 2.0 + 102.0 + wide_shift * 0.3
        button_base_pos = Vec2(screen_w / 2.0 + 128.0, button_base_y + 50.0)
        scale = 1.0
        button_w = button_width(
            small, self._purchase_button.label, scale=scale, force_wide=self._purchase_button.force_wide
        )
        button_draw(textures, small, self._purchase_button, pos=button_base_pos, width=button_w, scale=scale)
        button_draw(
            textures,
            small,
            self._maybe_later_button,
            pos=button_base_pos.offset(dy=40.0),
            width=button_w,
            scale=scale,
        )

        # Demo purchase screen uses menu-style cursor; draw it explicitly since the OS cursor is hidden.
        particles = cache.get_or_load("particles", "game/particles.jaz").texture
        cursor_tex = cache.get_or_load("ui_cursor", "ui/ui_cursor.jaz").texture
        mouse = rl.get_mouse_position()
        pulse_time = float(self._upsell_pulse_ms) * 0.001
        draw_menu_cursor(particles, cursor_tex, pos=Vec2.from_xy(mouse), pulse_time=pulse_time)

    def _ensure_cache(self) -> PaqTextureCache:
        cache = self._state.texture_cache
        if cache is not None:
            return cache
        entries = load_paq_entries(self._state.assets_dir)
        cache = PaqTextureCache(entries=entries, textures={})
        self._state.texture_cache = cache
        return cache

    def _demo_mode_start(self) -> None:
        index = self._demo_variant_index
        self._demo_variant_index = (index + 1) % DEMO_VARIANT_COUNT
        self._variant_index = index
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._purchase_active = False
        self._purchase_url_opened = False
        self._spawn_rng.srand(self._state.rng.randrange(0, 0x1_0000_0000))
        self._world.state.bonuses.weapon_power_up = 0.0
        if index == 0:
            self._apply_variant_ground(0)
            self._setup_variant_0()
        elif index == 1:
            self._apply_variant_ground(1)
            self._setup_variant_1()
        elif index == 2:
            self._apply_variant_ground(2)
            self._setup_variant_2()
        elif index == 3:
            self._apply_variant_ground(3)
            self._setup_variant_3()
        elif index == 4:
            self._apply_variant_ground(4)
            self._setup_variant_0()
        else:
            # demo_purchase_interstitial_begin
            self._begin_purchase_screen(DEMO_PURCHASE_INTERSTITIAL_LIMIT_MS, reset_timeline=True)

        # demo_purchase_screen_update increments demo_upsell_message_index when the
        # timeline resets (quest_spawn_timeline == 0) and the purchase screen is inactive.
        if (not self._purchase_active) and _DEMO_UPSELL_MESSAGES:
            self._upsell_message_index = (self._upsell_message_index + 1) % len(_DEMO_UPSELL_MESSAGES)

    def _setup_world_players(self, specs: list[tuple[Vec2, int]]) -> None:
        seed = int(self._state.rng.getrandbits(32))
        self._world.reset(seed=seed, player_count=len(specs))
        for idx, (pos, weapon_id) in enumerate(specs):
            if idx >= len(self._world.players):
                continue
            player = self._world.players[idx]
            player.pos = pos
            # Keep aim anchored to the spawn position so demo aim starts stable.
            player.aim = pos
            weapon_assign_player(player, int(weapon_id))
        self._demo_targets = [None] * len(self._world.players)

    def _apply_variant_ground(self, index: int) -> None:
        if index == 5:
            return
        terrain = {
            0: (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
            1: (
                "ter_q2_base",
                "ter_q2_tex1",
                "ter/ter_q2_base.jaz",
                "ter/ter_q2_tex1.jaz",
            ),
            2: (
                "ter_q3_base",
                "ter_q3_tex1",
                "ter/ter_q3_base.jaz",
                "ter/ter_q3_tex1.jaz",
            ),
            3: (
                "ter_q4_base",
                "ter_q4_tex1",
                "ter/ter_q4_base.jaz",
                "ter/ter_q4_tex1.jaz",
            ),
            4: (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
        }.get(
            index,
            (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
        )
        base_key, overlay_key, base_path, overlay_path = terrain
        self._world.set_terrain(
            base_key=base_key,
            overlay_key=overlay_key,
            base_path=base_path,
            overlay_path=overlay_path,
        )

    def _crand_mod(self, mod: int) -> int:
        if mod <= 0:
            return 0
        return int(self._crand.rand() % mod)

    def _spawn(self, spawn_id: int, pos: Vec2, *, heading: float = 0.0) -> None:
        self._world.creatures.spawn_template(
            int(spawn_id),
            pos,
            float(heading),
            self._spawn_rng,
            rand=self._spawn_rng.rand,
        )

    def _setup_variant_0(self) -> None:
        self._demo_time_limit_ms = 4000
        # demo_setup_variant_0 uses weapon_id=0x0B.
        weapon_id = 11
        self._setup_world_players(
            [
                (Vec2(448.0, 384.0), weapon_id),
                (Vec2(546.0, 654.0), weapon_id),
            ]
        )
        y = 256
        i = 0
        while y < 1696:
            col = i % 2
            self._spawn(0x38, Vec2(float((col + 2) * 64), float(y)), heading=RANDOM_HEADING_SENTINEL)
            self._spawn(0x38, Vec2(float(col * 64 + 798), float(y)), heading=RANDOM_HEADING_SENTINEL)
            y += 80
            i += 1

    def _setup_variant_1(self) -> None:
        self._demo_time_limit_ms = 5000
        # demo_setup_variant_1 uses weapon_id=0x05.
        weapon_id = 5
        self._setup_world_players(
            [
                (Vec2(490.0, 448.0), weapon_id),
                (Vec2(480.0, 576.0), weapon_id),
            ]
        )
        self._world.state.bonuses.weapon_power_up = 15.0
        for idx in range(20):
            x = float(self._crand_mod(200) + 32)
            y = float(self._crand_mod(899) + 64)
            self._spawn(0x34, Vec2(x, y), heading=RANDOM_HEADING_SENTINEL)
            if idx % 3 != 0:
                spawn_pos = Vec2(float(self._crand_mod(30) + 32), float(self._crand_mod(899) + 64))
                self._spawn(0x35, spawn_pos, heading=RANDOM_HEADING_SENTINEL)

    def _setup_variant_2(self) -> None:
        self._demo_time_limit_ms = 5000
        # demo_setup_variant_2 uses weapon_id=0x15.
        weapon_id = 21
        self._setup_world_players([(Vec2(512.0, 512.0), weapon_id)])
        y = 128
        i = 0
        while y < 848:
            col = i % 2
            self._spawn(0x41, Vec2(float(col * 64 + 32), float(y)), heading=RANDOM_HEADING_SENTINEL)
            self._spawn(0x41, Vec2(float((col + 2) * 64), float(y)), heading=RANDOM_HEADING_SENTINEL)
            self._spawn(0x41, Vec2(float(col * 64 - 64), float(y)), heading=RANDOM_HEADING_SENTINEL)
            self._spawn(0x41, Vec2(float((col + 12) * 64), float(y)), heading=RANDOM_HEADING_SENTINEL)
            y += 60
            i += 1

    def _setup_variant_3(self) -> None:
        self._demo_time_limit_ms = 4000
        # demo_setup_variant_3 uses weapon_id=0x12.
        weapon_id = 18
        self._setup_world_players([(Vec2(512.0, 512.0), weapon_id)])
        for idx in range(20):
            x = float(self._crand_mod(200) + 32)
            y = float(self._crand_mod(899) + 64)
            self._spawn(0x24, Vec2(x, y), heading=0.0)
            if idx % 3 != 0:
                spawn_pos = Vec2(float(self._crand_mod(30) + 32), float(self._crand_mod(899) + 64))
                self._spawn(0x25, spawn_pos, heading=0.0)

    def _draw_overlay(self) -> None:
        if getattr(self._state, "demo_enabled", False):
            self._draw_demo_upsell_overlay()
            return
        title = f"DEMO MODE  ({self._variant_index + 1}/{DEMO_VARIANT_COUNT})"
        hint = "Press any key / click to skip"
        remaining = max(0.0, float(self._demo_time_limit_ms - self._quest_spawn_timeline_ms) / 1000.0)
        weapons = ", ".join(f"P{p.index + 1}:{_weapon_name(p.weapon_id)}" for p in self._world.players)
        detail = f"{weapons}  —  next in {remaining:0.1f}s"
        rl.draw_text(title, 16, 12, 20, rl.Color(240, 240, 240, 255))
        rl.draw_text(detail, 16, 36, 16, rl.Color(180, 180, 190, 255))
        rl.draw_text(hint, 16, 56, 16, rl.Color(140, 140, 150, 255))

    def _ensure_upsell_font(self) -> GrimMonoFont:
        if self._upsell_font is not None:
            return self._upsell_font
        missing_assets: list[str] = []
        self._upsell_font = load_grim_mono_font(self._state.assets_dir, missing_assets)
        return self._upsell_font

    def _draw_demo_upsell_overlay(self) -> None:
        # Modeled after the shareware "Want more ..." overlay in demo_purchase_screen_update
        # (crimsonland.exe 0x0040B740), but without the purchase screen.
        if not _DEMO_UPSELL_MESSAGES:
            return

        font = self._ensure_upsell_font()
        msg = _DEMO_UPSELL_MESSAGES[self._upsell_message_index]

        timeline_ms = self._quest_spawn_timeline_ms
        limit_ms = self._demo_time_limit_ms
        var_2c = float(timeline_ms) * 0.016

        alpha = 1.0
        if var_2c < 20.0:
            alpha = var_2c * 0.05
        if timeline_ms > limit_ms - 500:
            alpha = float(limit_ms - timeline_ms) * 0.002
        alpha = clamp(alpha, 0.0, 1.0)

        scale = 0.8
        text_w = float(len(msg)) * 12.8

        text_x = 50.0
        text_y = var_2c + 50.0
        bg_x = 60.0
        bg_y = text_y - 4.0
        bar_x = 64.0
        bar_y = var_2c + 72.0

        bg_alpha = int(round(clamp(alpha * 0.5, 0.0, 1.0) * 255.0))
        bar_alpha = int(round(clamp(alpha * 0.8, 0.0, 1.0) * 255.0))
        txt_alpha = int(round(clamp(alpha, 0.0, 1.0) * 255.0))

        rl.draw_rectangle_rec(
            rl.Rectangle(bg_x, bg_y, text_w + 12.0, 30.0),
            rl.Color(0, 0, 0, bg_alpha),
        )

        progress = 0.0
        if limit_ms > 0:
            progress = clamp(float(timeline_ms) / float(limit_ms), 0.0, 1.0)
        rl.draw_rectangle_rec(
            rl.Rectangle(bar_x, bar_y, text_w * progress, 3.0),
            rl.Color(128, 26, 26, bar_alpha),
        )

        draw_grim_mono_text(font, msg, Vec2(text_x, text_y), scale, rl.Color(255, 255, 255, txt_alpha))

    def _update_world(self, dt: float) -> None:
        if not self._world.players:
            return
        inputs = self._build_demo_inputs(dt)
        self._world.update(dt, inputs=inputs, auto_pick_perks=False, game_mode=0, perk_progression_enabled=False)

    def _build_demo_inputs(self, dt: float) -> list[PlayerInput]:
        players = self._world.players
        creatures = self._world.creatures.entries
        if len(self._demo_targets) != len(players):
            self._demo_targets = [None] * len(players)
        center = Vec2(float(self._world.world_size) * 0.5, float(self._world.world_size) * 0.5)

        dt = float(dt)

        def _turn_towards_heading(cur: float, target: float) -> tuple[float, float]:
            cur = cur % math.tau
            target = target % math.tau
            delta = (target - cur + math.pi) % math.tau - math.pi
            diff = abs(delta)
            if diff <= 1e-9:
                return cur, 0.0
            step = dt * diff * 5.0
            cur = (cur + step) % math.tau if delta > 0.0 else (cur - step) % math.tau
            return cur, diff

        inputs: list[PlayerInput] = []
        for idx, player in enumerate(players):
            target_idx = self._select_demo_target(idx, player, creatures)
            target = None
            if target_idx is not None and 0 <= target_idx < len(creatures):
                candidate = creatures[target_idx]
                if candidate.active and candidate.hp > 0.0:
                    target = candidate

            # Aim: ease the aim point toward the target.
            aim = player.aim
            auto_fire = False
            if target is not None:
                target_pos = target.pos
                aim_dir, aim_dist = (target_pos - aim).normalized_with_length()
                if aim_dist >= 4.0:
                    step = aim_dist * 6.0 * dt
                    aim += aim_dir * step
                else:
                    aim = target_pos
                auto_fire = aim_dist < 128.0
            else:
                away_from_center, amag = (player.pos - center).normalized_with_length()
                if amag <= 1e-6:
                    away_from_center = Vec2(0.0, -1.0)
                aim = player.pos + away_from_center * 60.0

            # Movement:
            # - orbit center if no target
            # - chase target when near center
            # - return to center when too far
            if target is None:
                move_delta = (player.pos - center).rotated(math.pi / 2.0)
            else:
                center_dist = (player.pos - center).length()
                if center_dist <= 300.0:
                    move_delta = target.pos - player.pos
                else:
                    move_delta = center - player.pos

            desired_dir, desired_mag = move_delta.normalized_with_length()
            if desired_mag <= 1e-6:
                move = Vec2()
            else:
                desired_heading = desired_dir.to_heading()
                smoothed_heading, angle_diff = _turn_towards_heading(float(player.heading), desired_heading)
                move_mag = max(0.001, (math.pi - angle_diff) / math.pi)
                move = Vec2.from_heading(smoothed_heading) * move_mag

            inputs.append(
                PlayerInput(
                    move=move,
                    aim=aim,
                    fire_down=auto_fire,
                    fire_pressed=auto_fire,
                    reload_pressed=False,
                )
            )

        return inputs

    def _nearest_world_creature_index(self, pos: Vec2) -> int | None:
        best_idx = None
        best_dist = 0.0
        for idx, creature in enumerate(self._world.creatures.entries):
            if not (creature.active and creature.hp > 0.0):
                continue
            d = Vec2.distance_sq(pos, creature.pos)
            if best_idx is None or d < best_dist:
                best_idx = idx
                best_dist = d
        return best_idx

    def _select_demo_target(self, player_index: int, player: PlayerState, creatures: list) -> int | None:
        candidate = self._nearest_world_creature_index(player.pos)
        current = self._demo_targets[player_index] if player_index < len(self._demo_targets) else None
        if current is None:
            self._demo_targets[player_index] = candidate
            return candidate
        if not (0 <= current < len(creatures)):
            self._demo_targets[player_index] = candidate
            return candidate
        current_creature = creatures[current]
        if current_creature.hp <= 0.0 or not current_creature.active:
            self._demo_targets[player_index] = candidate
            return candidate
        if candidate is None or candidate == current:
            return current
        cand_creature = creatures[candidate]
        if not cand_creature.active or cand_creature.hp <= 0.0:
            return current
        cur_d = (current_creature.pos - player.pos).length()
        cand_d = (cand_creature.pos - player.pos).length()
        if cand_d + 64.0 < cur_d:
            self._demo_targets[player_index] = candidate
            return candidate
        return current



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
import math

import pyray as rl

from grim.config import CrimsonConfig, default_player_keybind_block
from grim.geom import Vec2

from .frontend.panels.controls_labels import controls_method_values
from .gameplay import PlayerInput, PlayerState
from .input_codes import (
    config_keybinds_for_player,
    input_axis_value_for_player,
    input_code_is_down_for_player,
    input_code_is_pressed_for_player,
)

_AIM_RADIUS_KEYBOARD = 60.0
_AIM_RADIUS_PAD_BASE = 42.0
_AIM_RADIUS_PAD_SCALE = 20.0
_POINT_CLICK_STOP_RADIUS = 20.0
_AIM_KEYBOARD_TURN_RATE = 3.0
_AIM_JOYSTICK_TURN_RATE = 4.0

_MOVE_SLOT_UP = 0
_MOVE_SLOT_DOWN = 1
_MOVE_SLOT_LEFT = 2
_MOVE_SLOT_RIGHT = 3
_FIRE_SLOT = 4
_AIM_LEFT_SLOT = 7
_AIM_RIGHT_SLOT = 8
_AIM_AXIS_Y_SLOT = 9
_AIM_AXIS_X_SLOT = 10
_MOVE_AXIS_Y_SLOT = 11
_MOVE_AXIS_X_SLOT = 12


@dataclass(slots=True)
class _PerPlayerInputState:
    aim_heading: float = 0.0
    move_target: Vec2 = field(default_factory=lambda: Vec2(-1.0, -1.0))


def _is_finite(v: float) -> bool:
    return math.isfinite(float(v))


def _clamp_unit(v: float) -> float:
    value = float(v)
    if value < -1.0:
        return -1.0
    if value > 1.0:
        return 1.0
    return value


def _aim_point_from_heading(pos: Vec2, heading: float, *, radius: float = _AIM_RADIUS_KEYBOARD) -> Vec2:
    return pos + Vec2.from_heading(float(heading)) * float(radius)


def _load_player_bind_block(config: CrimsonConfig | None, *, player_index: int) -> tuple[int, ...]:
    binds = config_keybinds_for_player(config, player_index=int(player_index))
    if len(binds) >= 16:
        return tuple(int(v) for v in binds[:16])
    return tuple(int(v) for v in default_player_keybind_block(int(player_index)))


def clear_input_edges(inputs: Sequence[PlayerInput]) -> list[PlayerInput]:
    return [
        PlayerInput(
            move=inp.move,
            aim=inp.aim,
            fire_down=bool(inp.fire_down),
            fire_pressed=False,
            reload_pressed=False,
            move_forward_pressed=inp.move_forward_pressed,
            move_backward_pressed=inp.move_backward_pressed,
            turn_left_pressed=inp.turn_left_pressed,
            turn_right_pressed=inp.turn_right_pressed,
        )
        for inp in inputs
    ]


class LocalInputInterpreter:
    def __init__(self) -> None:
        self._states: list[_PerPlayerInputState] = [_PerPlayerInputState() for _ in range(4)]

    def reset(self, *, players: Sequence[PlayerState] | None = None) -> None:
        for idx in range(4):
            state = self._states[idx]
            state.move_target = Vec2(-1.0, -1.0)
            heading = 0.0
            if players is not None and idx < len(players):
                candidate = float(getattr(players[idx], "aim_heading", 0.0))
                if _is_finite(candidate):
                    heading = candidate
            state.aim_heading = float(heading)

    def _state_for_player(self, player_index: int, *, player: PlayerState | None = None) -> _PerPlayerInputState:
        idx = max(0, min(3, int(player_index)))
        state = self._states[idx]
        if player is not None and (not _is_finite(state.aim_heading)):
            state.aim_heading = float(getattr(player, "aim_heading", 0.0) or 0.0)
        return state

    @staticmethod
    def _reload_key(config: CrimsonConfig | None) -> int:
        if config is None:
            return 0x102
        return int(config.data.get("keybind_reload", 0x102) or 0x102)

    @staticmethod
    def _safe_controls_modes(config: CrimsonConfig | None, *, player_index: int) -> tuple[int, int]:
        if config is None:
            return 0, 2
        aim_scheme, move_mode = controls_method_values(config.data, player_index=int(player_index))
        return int(aim_scheme), int(move_mode)

    def build_player_input(
        self,
        *,
        player_index: int,
        player: PlayerState,
        config: CrimsonConfig | None,
        mouse_screen: Vec2,
        mouse_world: Vec2,
        screen_center: Vec2,
        dt_frame: float,
    ) -> PlayerInput:
        idx = max(0, min(3, int(player_index)))
        state = self._state_for_player(idx, player=player)
        binds = _load_player_bind_block(config, player_index=idx)
        aim_scheme, move_mode = self._safe_controls_modes(config, player_index=idx)
        reload_key = self._reload_key(config)

        up_key = int(binds[_MOVE_SLOT_UP])
        down_key = int(binds[_MOVE_SLOT_DOWN])
        left_key = int(binds[_MOVE_SLOT_LEFT])
        right_key = int(binds[_MOVE_SLOT_RIGHT])
        fire_key = int(binds[_FIRE_SLOT])
        aim_left_key = int(binds[_AIM_LEFT_SLOT])
        aim_right_key = int(binds[_AIM_RIGHT_SLOT])
        aim_axis_y = int(binds[_AIM_AXIS_Y_SLOT])
        aim_axis_x = int(binds[_AIM_AXIS_X_SLOT])
        move_axis_y = int(binds[_MOVE_AXIS_Y_SLOT])
        move_axis_x = int(binds[_MOVE_AXIS_X_SLOT])

        move_vec = Vec2()
        move_forward_pressed: bool | None = None
        move_backward_pressed: bool | None = None
        turn_left_pressed: bool | None = None
        turn_right_pressed: bool | None = None

        if int(move_mode) == 1:
            move_forward_pressed = bool(input_code_is_down_for_player(up_key, player_index=idx))
            move_backward_pressed = bool(input_code_is_down_for_player(down_key, player_index=idx))
            turn_left_pressed = bool(input_code_is_down_for_player(left_key, player_index=idx))
            turn_right_pressed = bool(input_code_is_down_for_player(right_key, player_index=idx))
            move_vec = Vec2(
                float(turn_right_pressed) - float(turn_left_pressed),
                float(move_backward_pressed) - float(move_forward_pressed),
            )
        elif int(move_mode) == 3:
            axis_y = -input_axis_value_for_player(move_axis_y, player_index=idx)
            axis_x = -input_axis_value_for_player(move_axis_x, player_index=idx)
            move_vec = Vec2(_clamp_unit(axis_x), _clamp_unit(axis_y))
        elif int(move_mode) == 4:
            if input_code_is_down_for_player(reload_key, player_index=idx):
                state.move_target = mouse_world
            if float(state.move_target.x) >= 0.0 and float(state.move_target.y) >= 0.0:
                delta = state.move_target - player.pos
                _dir, dist = delta.normalized_with_length()
                if float(dist) > _POINT_CLICK_STOP_RADIUS:
                    move_vec = _dir
        elif int(move_mode) == 5:
            move_vec = Vec2()
        else:
            move_vec = Vec2(
                float(input_code_is_down_for_player(right_key, player_index=idx))
                - float(input_code_is_down_for_player(left_key, player_index=idx)),
                float(input_code_is_down_for_player(down_key, player_index=idx))
                - float(input_code_is_down_for_player(up_key, player_index=idx)),
            )

        heading = float(state.aim_heading)
        if not _is_finite(heading):
            heading = float(getattr(player, "aim_heading", 0.0) or 0.0)
        aim = _aim_point_from_heading(player.pos, heading)
        if int(aim_scheme) == 0:
            aim = mouse_world
            delta = aim - player.pos
            if delta.length_sq() > 1e-9:
                heading = delta.to_heading()
        elif int(aim_scheme) == 1:
            if int(move_mode) in {1, 2}:
                if input_code_is_down_for_player(aim_right_key, player_index=idx):
                    heading = float(heading + float(dt_frame) * _AIM_KEYBOARD_TURN_RATE)
                if input_code_is_down_for_player(aim_left_key, player_index=idx):
                    heading = float(heading - float(dt_frame) * _AIM_KEYBOARD_TURN_RATE)
            aim = _aim_point_from_heading(player.pos, heading)
        elif int(aim_scheme) == 3:
            rel = mouse_screen - screen_center
            if rel.length_sq() > 1.0:
                heading = rel.to_heading()
            aim = _aim_point_from_heading(player.pos, heading)
        elif int(aim_scheme) == 4:
            axis_y = input_axis_value_for_player(aim_axis_y, player_index=idx)
            axis_x = input_axis_value_for_player(aim_axis_x, player_index=idx)
            axis_vec = Vec2(axis_x, axis_y)
            mag_sq = axis_vec.length_sq()
            if mag_sq > 1e-9:
                axis_dir, mag = axis_vec.normalized_with_length()
                heading = axis_dir.to_heading()
                radius = _AIM_RADIUS_PAD_BASE + mag * _AIM_RADIUS_PAD_SCALE
                aim = player.pos + axis_dir * radius
            else:
                aim = _aim_point_from_heading(player.pos, heading)
        elif int(aim_scheme) == 2:
            if input_code_is_down_for_player(aim_right_key, player_index=idx):
                heading = float(heading + float(dt_frame) * _AIM_JOYSTICK_TURN_RATE)
            if input_code_is_down_for_player(aim_left_key, player_index=idx):
                heading = float(heading - float(dt_frame) * _AIM_JOYSTICK_TURN_RATE)
            aim = _aim_point_from_heading(player.pos, heading)
        elif int(aim_scheme) == 5:
            aim = _aim_point_from_heading(player.pos, heading)

        state.aim_heading = float(heading)

        fire_down = bool(input_code_is_down_for_player(fire_key, player_index=idx))
        fire_pressed = bool(input_code_is_pressed_for_player(fire_key, player_index=idx))
        reload_pressed = bool(input_code_is_pressed_for_player(reload_key, player_index=idx)) if idx == 0 else False

        return PlayerInput(
            move=move_vec,
            aim=aim,
            fire_down=fire_down,
            fire_pressed=fire_pressed,
            reload_pressed=reload_pressed,
            move_forward_pressed=move_forward_pressed,
            move_backward_pressed=move_backward_pressed,
            turn_left_pressed=turn_left_pressed,
            turn_right_pressed=turn_right_pressed,
        )

    def build_frame_inputs(
        self,
        *,
        players: Sequence[PlayerState],
        config: CrimsonConfig | None,
        mouse_screen: Vec2,
        screen_to_world: Callable[[Vec2], Vec2],
        dt_frame: float,
    ) -> list[PlayerInput]:
        mouse_world = screen_to_world(mouse_screen)
        screen_center = Vec2(float(rl.get_screen_width()) * 0.5, float(rl.get_screen_height()) * 0.5)
        out: list[PlayerInput] = []
        for idx, player in enumerate(players):
            out.append(
                self.build_player_input(
                    player_index=idx,
                    player=player,
                    config=config,
                    mouse_screen=mouse_screen,
                    mouse_world=mouse_world,
                    screen_center=screen_center,
                    dt_frame=float(dt_frame),
                )
            )
        return out



from __future__ import annotations

"""Perk-specific deterministic hooks used by the world-step pipeline."""

from .registry import PLAYER_DEATH_HOOKS, WORLD_DT_STEPS

__all__ = ["PLAYER_DEATH_HOOKS", "WORLD_DT_STEPS"]



from __future__ import annotations

from dataclasses import dataclass
import math
from typing import Callable

from grim.geom import Vec2

from ..gameplay import PlayerState, perk_active
from ..perks import PerkId
from .runtime import CREATURE_HITBOX_ALIVE, CreatureState
from .spawn import CreatureFlags


def _owner_id_to_player_index(owner_id: int) -> int | None:
    if owner_id == -100:
        return 0
    if owner_id < 0:
        return -1 - owner_id
    return None


@dataclass(slots=True)
class _CreatureDamageCtx:
    creature: CreatureState
    damage: float
    damage_type: int
    impulse: Vec2
    owner_id: int
    dt: float
    players: list[PlayerState]
    rand: Callable[[], int]
    attacker: PlayerState | None


_CreatureDamageStep = Callable[[_CreatureDamageCtx], None]


def _damage_type1_uranium_filled_bullets(ctx: _CreatureDamageCtx) -> None:
    if ctx.attacker is None or not perk_active(ctx.attacker, PerkId.URANIUM_FILLED_BULLETS):
        return
    ctx.damage *= 2.0


def _damage_type1_living_fortress(ctx: _CreatureDamageCtx) -> None:
    attacker = ctx.attacker
    if attacker is None or not perk_active(attacker, PerkId.LIVING_FORTRESS):
        return
    for player in ctx.players:
        if float(player.health) <= 0.0:
            continue
        timer = float(player.living_fortress_timer)
        if timer > 0.0:
            ctx.damage *= timer * 0.05 + 1.0


def _damage_type1_barrel_greaser(ctx: _CreatureDamageCtx) -> None:
    if ctx.attacker is None or not perk_active(ctx.attacker, PerkId.BARREL_GREASER):
        return
    ctx.damage *= 1.4


def _damage_type1_doctor(ctx: _CreatureDamageCtx) -> None:
    if ctx.attacker is None or not perk_active(ctx.attacker, PerkId.DOCTOR):
        return
    ctx.damage *= 1.2


def _damage_type1_heading_jitter(ctx: _CreatureDamageCtx) -> None:
    creature = ctx.creature
    if (creature.flags & CreatureFlags.ANIM_PING_PONG) != 0:
        return
    jitter = float((int(ctx.rand()) & 0x7F) - 0x40) * 0.002
    size = max(1e-6, float(creature.size))
    turn = jitter / (size * 0.025)
    turn = min(math.pi / 2.0, turn)
    creature.heading += turn


def _damage_type7_ion_gun_master(ctx: _CreatureDamageCtx) -> None:
    if any(perk_active(player, PerkId.ION_GUN_MASTER) for player in ctx.players):
        ctx.damage *= 1.2


def _damage_type4_pyromaniac(ctx: _CreatureDamageCtx) -> None:
    if ctx.attacker is None or not perk_active(ctx.attacker, PerkId.PYROMANIAC):
        return
    ctx.damage *= 1.5
    ctx.rand()


_CREATURE_DAMAGE_ATTACKER_PRE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    1: (
        _damage_type1_uranium_filled_bullets,
        _damage_type1_living_fortress,
        _damage_type1_barrel_greaser,
        _damage_type1_doctor,
    ),
}

_CREATURE_DAMAGE_GLOBAL_PRE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    7: (_damage_type7_ion_gun_master,),
}


_CREATURE_DAMAGE_ATTACKER_ALIVE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    4: (_damage_type4_pyromaniac,),
}


def creature_apply_damage(
    creature: CreatureState,
    *,
    damage_amount: float,
    damage_type: int,
    impulse: Vec2,
    owner_id: int,
    dt: float,
    players: list[PlayerState],
    rand: Callable[[], int],
) -> bool:
    """Apply damage to a creature, returning True if the hit killed it.

    This is a partial port of `creature_apply_damage` (FUN_004207c0).

    Notes:
    - Death side-effects are handled by the caller (see Phase 2 in `plan.md`).
    - `damage_type` is a native integer category; call sites must supply it.
    """

    creature.last_hit_owner_id = int(owner_id)
    creature.hit_flash_timer = 0.2

    player_index = _owner_id_to_player_index(owner_id)
    attacker = players[player_index] if player_index is not None and 0 <= player_index < len(players) else None

    ctx = _CreatureDamageCtx(
        creature=creature,
        damage=float(damage_amount),
        damage_type=int(damage_type),
        impulse=impulse,
        owner_id=int(owner_id),
        dt=float(dt),
        players=players,
        rand=rand,
        attacker=attacker,
    )

    for step in _CREATURE_DAMAGE_GLOBAL_PRE_STEPS.get(ctx.damage_type, ()):
        step(ctx)

    if attacker is not None:
        for step in _CREATURE_DAMAGE_ATTACKER_PRE_STEPS.get(ctx.damage_type, ()):
            step(ctx)
    if ctx.damage_type == 1:
        _damage_type1_heading_jitter(ctx)

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.hitbox_size -= float(dt) * 15.0
        return True

    if attacker is not None:
        for step in _CREATURE_DAMAGE_ATTACKER_ALIVE_STEPS.get(ctx.damage_type, ()):
            step(ctx)

    creature.hp -= float(ctx.damage)
    creature.vel = creature.vel - ctx.impulse

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.hitbox_size = float(creature.hitbox_size) - float(dt)
        else:
            creature.hitbox_size = float(creature.hitbox_size) - 0.001
        creature.vel = creature.vel - impulse * 2.0
        return True

    if creature.hitbox_size != CREATURE_HITBOX_ALIVE and dt > 0.0:
        creature.hitbox_size = CREATURE_HITBOX_ALIVE

    return False



from __future__ import annotations

from .types import QuestContext, QuestDefinition, SpawnEntry
from .registry import all_quests, quest_by_level
from . import tier1, tier2, tier3, tier4, tier5

__all__ = [
    "QuestContext",
    "QuestDefinition",
    "SpawnEntry",
    "all_quests",
    "quest_by_level",
    "tier1",
    "tier2",
    "tier3",
    "tier4",
    "tier5",
]



from __future__ import annotations

"""
Weapon definitions for the rewrite runtime.

This file is **manually maintained** (do not auto-generate).

It was originally seeded from `weapon_table_init` (`FUN_004519b0`) and the
rewrite now uses the **native 1-based weapon ids** (e.g. pistol is
`weapon_id=1`). In the native code, projectile `type_id` values passed into
`projectile_spawn` are **weapon table indices** (same numeric domain as weapon
ids). They are **not** a 1:1 mapping: multiple weapons can share a projectile
type id (e.g. Sawed-off and Jackhammer use the Shotgun template), and some
weapons bypass `projectile_spawn` entirely (particles / secondary pool).

Use `projectile_type_id_from_weapon_id` for the primary projectile `type_id`
(or `None` for non-projectile weapons), and `projectile_type_ids_from_weapon_id`
when you need the full set.

Reference material:
- `docs/weapon-table.md` (native struct + fields)
- `docs/weapon-id-map.md` (native ids + names)
"""

from dataclasses import dataclass
from enum import IntEnum

MANUALLY_MAINTAINED = True


class WeaponId(IntEnum):
    NONE = 0
    PISTOL = 1
    ASSAULT_RIFLE = 2
    SHOTGUN = 3
    SAWED_OFF_SHOTGUN = 4
    SUBMACHINE_GUN = 5
    GAUSS_GUN = 6
    MEAN_MINIGUN = 7
    FLAMETHROWER = 8
    PLASMA_RIFLE = 9
    MULTI_PLASMA = 10
    PLASMA_MINIGUN = 11
    ROCKET_LAUNCHER = 12
    SEEKER_ROCKETS = 13
    PLASMA_SHOTGUN = 14
    BLOW_TORCH = 15
    HR_FLAMER = 16
    MINI_ROCKET_SWARMERS = 17
    ROCKET_MINIGUN = 18
    PULSE_GUN = 19
    JACKHAMMER = 20
    ION_RIFLE = 21
    ION_MINIGUN = 22
    ION_CANNON = 23
    SHRINKIFIER_5K = 24
    BLADE_GUN = 25
    SPIDER_PLASMA = 26
    EVIL_SCYTHE = 27
    PLASMA_CANNON = 28
    SPLITTER_GUN = 29
    GAUSS_SHOTGUN = 30
    ION_SHOTGUN = 31
    FLAMEBURST = 32
    RAYGUN = 33
    UNKNOWN_34 = 34
    UNKNOWN_35 = 35
    UNKNOWN_36 = 36
    UNKNOWN_37 = 37
    UNKNOWN_38 = 38
    UNKNOWN_39 = 39
    UNKNOWN_40 = 40
    PLAGUE_SPHREADER_GUN = 41
    BUBBLEGUN = 42
    RAINBOW_GUN = 43
    GRIM_WEAPON = 44
    FIRE_BULLETS = 45
    UNKNOWN_46 = 46
    UNKNOWN_47 = 47
    UNKNOWN_48 = 48
    UNKNOWN_49 = 49
    TRANSMUTATOR = 50
    BLASTER_R_300 = 51
    LIGHTING_RIFLE = 52
    NUKE_LAUNCHER = 53


@dataclass(frozen=True)
class Weapon:
    weapon_id: int
    name: str | None
    ammo_class: int | None
    clip_size: int | None
    shot_cooldown: float | None
    reload_time: float | None
    spread_heat_inc: float | None
    fire_sound: str | None
    reload_sound: str | None
    icon_index: int | None
    flags: int | None
    projectile_meta: int | None
    damage_scale: float | None
    pellet_count: int | None


WEAPON_TABLE = [
    Weapon(
        weapon_id=1,
        name='Pistol',
        ammo_class=0,
        clip_size=10,
        shot_cooldown=0.7117,
        reload_time=1.2,
        spread_heat_inc=0.22,
        fire_sound='sfx_pistol_fire',
        reload_sound='sfx_pistol_reload',
        icon_index=0,
        flags=5,
        projectile_meta=55,
        damage_scale=4.1,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=2,
        name='Assault Rifle',
        ammo_class=0,
        clip_size=25,
        shot_cooldown=0.117,
        reload_time=1.2,
        spread_heat_inc=0.09,
        fire_sound='sfx_autorifle_fire',
        reload_sound='sfx_autorifle_reload',
        icon_index=1,
        flags=1,
        projectile_meta=50,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=3,
        name='Shotgun',
        ammo_class=0,
        clip_size=12,
        shot_cooldown=0.85,
        reload_time=1.9,
        spread_heat_inc=0.27,
        fire_sound='sfx_shotgun_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=2,
        flags=1,
        projectile_meta=60,
        damage_scale=1.2,
        pellet_count=12,
    ),
    Weapon(
        weapon_id=4,
        name='Sawed-off Shotgun',
        ammo_class=0,
        clip_size=12,
        shot_cooldown=0.87,
        reload_time=1.9,
        spread_heat_inc=0.13,
        fire_sound='_DAT_004d8434',
        reload_sound='_DAT_004d93bc',
        icon_index=3,
        flags=1,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=12,
    ),
    Weapon(
        weapon_id=5,
        name='Submachine Gun',
        ammo_class=0,
        clip_size=30,
        shot_cooldown=0.088117,
        reload_time=1.2,
        spread_heat_inc=0.082,
        fire_sound='sfx_hrpm_fire',
        reload_sound='_DAT_004d83c0',
        icon_index=4,
        flags=5,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=6,
        name='Gauss Gun',
        ammo_class=0,
        clip_size=6,
        shot_cooldown=0.6,
        reload_time=1.6,
        spread_heat_inc=0.42,
        fire_sound='sfx_gauss_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=5,
        flags=1,
        projectile_meta=215,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=7,
        name='Mean Minigun',
        ammo_class=0,
        clip_size=120,
        shot_cooldown=0.09,
        reload_time=4.0,
        spread_heat_inc=0.062,
        fire_sound='sfx_autorifle_fire',
        reload_sound='_DAT_004d83c0',
        icon_index=6,
        flags=3,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=8,
        name='Flamethrower',
        ammo_class=1,
        clip_size=30,
        shot_cooldown=0.008113,
        reload_time=2.0,
        spread_heat_inc=0.015,
        fire_sound='sfx_flamer_fire_01',
        reload_sound='_DAT_004d83c0',
        icon_index=7,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=9,
        name='Plasma Rifle',
        ammo_class=0,
        clip_size=20,
        shot_cooldown=0.2908117,
        reload_time=1.2,
        spread_heat_inc=0.182,
        fire_sound='sfx_shock_fire',
        reload_sound='_DAT_004d83c0',
        icon_index=8,
        flags=None,
        projectile_meta=30,
        damage_scale=5.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=10,
        name='Multi-Plasma',
        ammo_class=0,
        clip_size=8,
        shot_cooldown=0.6208117,
        reload_time=1.4,
        spread_heat_inc=0.32,
        fire_sound='sfx_shock_fire',
        reload_sound='_DAT_004d83c0',
        icon_index=9,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=3,
    ),
    Weapon(
        weapon_id=11,
        name='Plasma Minigun',
        ammo_class=0,
        clip_size=30,
        shot_cooldown=0.11,
        reload_time=1.3,
        spread_heat_inc=0.097,
        fire_sound='sfx_plasmaminigun_fire',
        reload_sound='_DAT_004d83c0',
        icon_index=10,
        flags=None,
        projectile_meta=35,
        damage_scale=2.1,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=12,
        name='Rocket Launcher',
        ammo_class=2,
        clip_size=5,
        shot_cooldown=0.7408117,
        reload_time=1.2,
        spread_heat_inc=0.42,
        fire_sound='sfx_rocket_fire',
        reload_sound='sfx_autorifle_reload_alt',
        icon_index=11,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=13,
        name='Seeker Rockets',
        ammo_class=2,
        clip_size=8,
        shot_cooldown=0.3108117,
        reload_time=1.2,
        spread_heat_inc=0.32,
        fire_sound='sfx_rocket_fire',
        reload_sound='sfx_autorifle_reload_alt',
        icon_index=12,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=14,
        name='Plasma Shotgun',
        ammo_class=0,
        clip_size=8,
        shot_cooldown=0.48,
        reload_time=3.1,
        spread_heat_inc=0.11,
        fire_sound='sfx_plasmashotgun_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=13,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=14,
    ),
    Weapon(
        weapon_id=15,
        name='Blow Torch',
        ammo_class=1,
        clip_size=30,
        shot_cooldown=0.006113,
        reload_time=1.5,
        spread_heat_inc=0.01,
        fire_sound='sfx_flamer_fire_01',
        reload_sound='_DAT_004d83c0',
        icon_index=14,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=16,
        name='HR Flamer',
        ammo_class=1,
        clip_size=30,
        shot_cooldown=0.0085,
        reload_time=1.8,
        spread_heat_inc=0.01,
        fire_sound='sfx_flamer_fire_01',
        reload_sound='_DAT_004d83c0',
        icon_index=15,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=17,
        name='Mini-Rocket Swarmers',
        ammo_class=2,
        clip_size=5,
        shot_cooldown=1.8,
        reload_time=1.8,
        spread_heat_inc=0.12,
        fire_sound='sfx_rocket_fire',
        reload_sound='sfx_autorifle_reload_alt',
        icon_index=16,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=18,
        name='Rocket Minigun',
        ammo_class=2,
        clip_size=16,
        shot_cooldown=0.12,
        reload_time=1.8,
        spread_heat_inc=0.12,
        fire_sound='sfx_rocketmini_fire',
        reload_sound='sfx_autorifle_reload_alt',
        icon_index=17,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=19,
        name='Pulse Gun',
        ammo_class=3,
        clip_size=16,
        shot_cooldown=0.1,
        reload_time=0.1,
        spread_heat_inc=0.0,
        fire_sound='sfx_pulse_fire',
        reload_sound='sfx_autorifle_reload',
        icon_index=18,
        flags=8,
        projectile_meta=20,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=20,
        name='Jackhammer',
        ammo_class=0,
        clip_size=16,
        shot_cooldown=0.14,
        reload_time=3.0,
        spread_heat_inc=0.16,
        fire_sound='sfx_shotgun_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=19,
        flags=1,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=4,
    ),
    Weapon(
        weapon_id=21,
        name='Ion Rifle',
        ammo_class=4,
        clip_size=8,
        shot_cooldown=0.4,
        reload_time=1.35,
        spread_heat_inc=0.112,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=20,
        flags=8,
        projectile_meta=15,
        damage_scale=3.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=22,
        name='Ion Minigun',
        ammo_class=4,
        clip_size=20,
        shot_cooldown=0.1,
        reload_time=1.8,
        spread_heat_inc=0.09,
        fire_sound='sfx_shockminigun_fire',
        reload_sound='_DAT_004d86a8',
        icon_index=21,
        flags=8,
        projectile_meta=20,
        damage_scale=1.4,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=23,
        name='Ion Cannon',
        ammo_class=4,
        clip_size=3,
        shot_cooldown=1.0,
        reload_time=3.0,
        spread_heat_inc=0.68,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=22,
        flags=None,
        projectile_meta=10,
        damage_scale=16.7,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=24,
        name='Shrinkifier 5k',
        ammo_class=0,
        clip_size=8,
        shot_cooldown=0.21,
        reload_time=1.22,
        spread_heat_inc=0.04,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=23,
        flags=8,
        projectile_meta=45,
        damage_scale=0.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=25,
        name='Blade Gun',
        ammo_class=0,
        clip_size=6,
        shot_cooldown=0.35,
        reload_time=3.5,
        spread_heat_inc=0.04,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='sfx_shock_reload',
        icon_index=24,
        flags=8,
        projectile_meta=20,
        damage_scale=11.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=26,
        name='Spider Plasma',
        ammo_class=0,
        clip_size=5,
        shot_cooldown=0.2,
        reload_time=1.2,
        spread_heat_inc=0.04,
        fire_sound='_DAT_004d92bc',
        reload_sound='_DAT_004d93bc',
        icon_index=25,
        flags=8,
        projectile_meta=10,
        damage_scale=0.5,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=27,
        name='Evil Scythe',
        ammo_class=4,
        clip_size=3,
        shot_cooldown=1.0,
        reload_time=3.0,
        spread_heat_inc=0.68,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=25,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=28,
        name='Plasma Cannon',
        ammo_class=0,
        clip_size=3,
        shot_cooldown=0.9,
        reload_time=2.7,
        spread_heat_inc=0.6,
        fire_sound='sfx_shock_fire',
        reload_sound='_DAT_004d86a8',
        icon_index=25,
        flags=None,
        projectile_meta=10,
        damage_scale=28.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=29,
        name='Splitter Gun',
        ammo_class=0,
        clip_size=6,
        shot_cooldown=0.7,
        reload_time=2.2,
        spread_heat_inc=0.28,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=28,
        flags=None,
        projectile_meta=30,
        damage_scale=6.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=30,
        name='Gauss Shotgun',
        ammo_class=0,
        clip_size=4,
        shot_cooldown=1.05,
        reload_time=2.1,
        spread_heat_inc=0.27,
        fire_sound='sfx_gauss_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=30,
        flags=1,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=31,
        name='Ion Shotgun',
        ammo_class=4,
        clip_size=10,
        shot_cooldown=0.85,
        reload_time=1.9,
        spread_heat_inc=0.27,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=31,
        flags=1,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=8,
    ),
    Weapon(
        weapon_id=32,
        name='Flameburst',
        ammo_class=4,
        clip_size=60,
        shot_cooldown=0.02,
        reload_time=3.0,
        spread_heat_inc=0.18,
        fire_sound='sfx_flamer_fire_01',
        reload_sound='_DAT_004d86a8',
        icon_index=29,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=33,
        name='RayGun',
        ammo_class=4,
        clip_size=12,
        shot_cooldown=0.7,
        reload_time=2.0,
        spread_heat_inc=0.38,
        fire_sound='sfx_shock_fire_alt',
        reload_sound='_DAT_004d86a8',
        icon_index=30,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=41,
        name='Plague Sphreader Gun',
        ammo_class=None,
        clip_size=5,
        shot_cooldown=0.2,
        reload_time=1.2,
        spread_heat_inc=0.04,
        fire_sound='sfx_bloodspill_01',
        reload_sound='_DAT_004d93bc',
        icon_index=40,
        flags=8,
        projectile_meta=15,
        damage_scale=0.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=42,
        name='Bubblegun',
        ammo_class=None,
        clip_size=15,
        shot_cooldown=0.1613,
        reload_time=1.2,
        spread_heat_inc=0.05,
        fire_sound='_DAT_004d92bc',
        reload_sound='_DAT_004d93bc',
        icon_index=41,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=43,
        name='Rainbow Gun',
        ammo_class=None,
        clip_size=10,
        shot_cooldown=0.2,
        reload_time=1.2,
        spread_heat_inc=0.09,
        fire_sound='_DAT_004d92bc',
        reload_sound='_DAT_004d93bc',
        icon_index=42,
        flags=8,
        projectile_meta=10,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=44,
        name='Grim Weapon',
        ammo_class=None,
        clip_size=3,
        shot_cooldown=0.5,
        reload_time=1.2,
        spread_heat_inc=0.4,
        fire_sound='_DAT_004d92bc',
        reload_sound='_DAT_004d93bc',
        icon_index=43,
        flags=None,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=45,
        name='Fire bullets',
        ammo_class=None,
        clip_size=112,
        shot_cooldown=0.14,
        reload_time=1.2,
        spread_heat_inc=0.22,
        fire_sound='_DAT_004d7b7c',
        reload_sound='sfx_pistol_reload',
        icon_index=44,
        flags=1,
        projectile_meta=60,
        damage_scale=0.25,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=50,
        name='Transmutator',
        ammo_class=None,
        clip_size=50,
        shot_cooldown=0.04,
        reload_time=5.0,
        spread_heat_inc=0.04,
        fire_sound='sfx_bloodspill_01',
        reload_sound='_DAT_004d93bc',
        icon_index=49,
        flags=9,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=51,
        name='Blaster R-300',
        ammo_class=None,
        clip_size=20,
        shot_cooldown=0.08,
        reload_time=2.0,
        spread_heat_inc=0.05,
        fire_sound='sfx_shock_fire',
        reload_sound='_DAT_004d93bc',
        icon_index=50,
        flags=9,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=52,
        name='Lighting Rifle',
        ammo_class=None,
        clip_size=500,
        shot_cooldown=4.0,
        reload_time=8.0,
        spread_heat_inc=1.0,
        fire_sound='sfx_explosion_large',
        reload_sound='sfx_shotgun_reload',
        icon_index=51,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
    Weapon(
        weapon_id=53,
        name='Nuke Launcher',
        ammo_class=None,
        clip_size=1,
        shot_cooldown=4.0,
        reload_time=8.0,
        spread_heat_inc=1.0,
        fire_sound='_DAT_004d93b4',
        reload_sound='_DAT_004d93bc',
        icon_index=52,
        flags=8,
        projectile_meta=45,
        damage_scale=1.0,
        pellet_count=1,
    ),
]

WEAPON_BY_ID = {
    entry.weapon_id: entry for entry in WEAPON_TABLE
}

WEAPON_PROJECTILE_TYPE_IDS: dict[int, tuple[int, ...]] = {
    # Source: analysis/ghidra/raw/crimsonland.exe_decompiled.c (`player_fire_weapon`).
    # Weapon ids not listed here use `type_id == weapon_id` in the native
    # `projectile_spawn` path.
    1: (0x01,),  # Pistol
    2: (0x02,),  # Assault Rifle
    3: (0x03,),  # Shotgun
    4: (0x03,),  # Sawed-off Shotgun
    5: (0x05,),  # Submachine Gun
    6: (0x06,),  # Gauss Gun
    7: (0x01,),  # Mean Minigun
    8: (),  # Flamethrower (particle path)
    9: (0x09,),  # Plasma Rifle
    10: (0x09, 0x0B),  # Multi-Plasma (spread includes 0x0B)
    11: (0x0B,),  # Plasma Minigun
    12: (),  # Rocket Launcher (secondary projectile pool)
    13: (),  # Seeker Rockets (secondary projectile pool)
    14: (0x0B,),  # Plasma Shotgun
    15: (),  # Blow Torch (particle path)
    16: (),  # HR Flamer (particle path)
    17: (),  # Mini-Rocket Swarmers (secondary projectile pool)
    18: (),  # Rocket Minigun (secondary projectile pool)
    19: (0x13,),  # Pulse Gun
    20: (0x03,),  # Jackhammer
    21: (0x15,),  # Ion Rifle
    22: (0x16,),  # Ion Minigun
    23: (0x17,),  # Ion Cannon
    24: (0x18,),  # Shrinkifier 5k
    25: (0x19,),  # Blade Gun
    28: (0x1C,),  # Plasma Cannon
    29: (0x1D,),  # Splitter Gun
    30: (0x06,),  # Gauss Shotgun
    31: (0x16,),  # Ion Shotgun
    41: (0x29,),  # Plague Spreader Gun
    42: (),  # Bubblegun (particle slow)
    43: (0x2B,),  # Rainbow Gun
    45: (0x2D,),  # Fire Bullets
}

def weapon_entry_for_projectile_type_id(type_id: int) -> Weapon | None:
    # Native `projectile_spawn` indexes the weapon table by `type_id`.
    return WEAPON_BY_ID.get(int(type_id))


def projectile_type_id_from_weapon_id(weapon_id: int) -> int | None:
    """Return the primary projectile `type_id` used by `weapon_id`.

    Returns `None` for weapons that don't use the main projectile pool.
    """

    weapon_id = int(weapon_id)
    type_ids = WEAPON_PROJECTILE_TYPE_IDS.get(weapon_id)
    if type_ids is not None:
        return int(type_ids[0]) if type_ids else None

    # Default native behavior for projectile weapons is `type_id == weapon_id`.
    if weapon_id in WEAPON_BY_ID:
        return weapon_id
    return None


def projectile_type_ids_from_weapon_id(weapon_id: int) -> tuple[int, ...]:
    """Return all projectile `type_id` values produced by `weapon_id`."""

    weapon_id = int(weapon_id)
    type_ids = WEAPON_PROJECTILE_TYPE_IDS.get(weapon_id)
    if type_ids is not None:
        return tuple(int(v) for v in type_ids)
    if weapon_id in WEAPON_BY_ID:
        return (weapon_id,)
    return ()


WEAPON_BY_NAME = {
    entry.name: entry for entry in WEAPON_TABLE if entry.name is not None
}



from __future__ import annotations

from dataclasses import dataclass
import math
import random
from typing import Iterator

from grim.geom import Vec2

from ..creatures.spawn import SpawnId
from .types import SpawnEntry


@dataclass(frozen=True, slots=True)
class EdgePoints:
    left: Vec2
    right: Vec2
    top: Vec2
    bottom: Vec2


def center_point(width: float, height: float | None = None) -> Vec2:
    if height is None:
        height = width
    return Vec2(float(width) * 0.5, float(height) * 0.5)


def edge_midpoints(width: float, height: float | None = None, offset: float = 64.0) -> EdgePoints:
    if height is None:
        height = width
    center = center_point(width, height)
    return EdgePoints(
        left=Vec2(-offset, center.y),
        right=Vec2(float(width) + offset, center.y),
        top=Vec2(center.x, -offset),
        bottom=Vec2(center.x, float(height) + offset),
    )


def corner_points(width: float, height: float | None = None, offset: float = 64.0) -> tuple[Vec2, ...]:
    if height is None:
        height = width
    return (
        Vec2(-offset, -offset),
        Vec2(float(width) + offset, -offset),
        Vec2(-offset, float(height) + offset),
        Vec2(float(width) + offset, float(height) + offset),
    )


def iter_angles(count: int, *, step: float | None = None, start: float = 0.0) -> Iterator[float]:
    if count <= 0:
        return iter(())
    if step is None:
        step = math.tau / float(count)
    for idx in range(count):
        yield start + float(idx) * step


def ring_points(
    center: Vec2,
    radius: float,
    count: int,
    *,
    step: float | None = None,
    start: float = 0.0,
) -> Iterator[tuple[Vec2, float]]:
    for angle in iter_angles(count, step=step, start=start):
        yield center + Vec2.from_angle(angle) * radius, angle


def random_angle(rng: random.Random) -> float:
    return float(rng.randrange(0x264)) * 0.01


def radial_points(
    center: Vec2,
    angle: float,
    radius_start: float,
    radius_end: float,
    radius_step: float,
) -> Iterator[Vec2]:
    direction = Vec2.from_angle(angle)
    radius = radius_start
    while radius < radius_end:
        yield center + direction * radius
        radius += radius_step


def heading_from_center(point: Vec2, center: Vec2) -> float:
    return (point - center).to_angle() - (math.pi / 2.0)


def line_points(start: Vec2, step: Vec2, count: int) -> Iterator[Vec2]:
    for idx in range(count):
        yield start + step * float(idx)


def spawn(
    point: Vec2,
    *,
    heading: float = 0.0,
    spawn_id: SpawnId,
    trigger_ms: int,
    count: int,
) -> SpawnEntry:
    return SpawnEntry(
        pos=point,
        heading=heading,
        spawn_id=spawn_id,
        trigger_ms=trigger_ms,
        count=count,
    )


def spawn_at(
    point: Vec2,
    *,
    heading: float = 0.0,
    spawn_id: SpawnId,
    trigger_ms: int,
    count: int,
) -> SpawnEntry:
    return spawn(
        point,
        heading=heading,
        spawn_id=spawn_id,
        trigger_ms=trigger_ms,
        count=count,
    )



from __future__ import annotations

from ...gameplay import PlayerState, perk_active
from ...perks import PerkId


def apply_reflex_boosted_dt(*, dt: float, players: list[PlayerState]) -> float:
    """Apply Reflex Boosted dt scaling from perk effects."""
    if float(dt) <= 0.0:
        return float(dt)
    if not players:
        return float(dt)
    if not perk_active(players[0], PerkId.REFLEX_BOOSTED):
        return float(dt)
    return float(dt) * 0.9



from __future__ import annotations
from grim.color import RGBA
from grim.geom import Vec2

from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Iterable

import math
import pyray as rl

from . import paq
from grim.math import clamp
from grim.fonts.grim_mono import (
    GrimMonoFont,
    draw_grim_mono_text,
    load_grim_mono_font,
)
from grim.fonts.small import (
    SmallFontData,
    draw_small_text,
    load_small_font,
    measure_small_text_width,
)

CONSOLE_LOG_NAME = "console.log"
MAX_CONSOLE_LINES = 0x1000
MAX_CONSOLE_INPUT = 0x3FF
DEFAULT_CONSOLE_HEIGHT = 300
EXTENDED_CONSOLE_HEIGHT = 480
CONSOLE_VERSION_TEXT = "Crimsonland 1.9.93"
CONSOLE_ANIM_SPEED = 3.5
CONSOLE_BLINK_SPEED = 3.0
CONSOLE_LINE_HEIGHT = 16.0
CONSOLE_MONO_SCALE = 0.5
CONSOLE_SMALL_SCALE = 1.0
CONSOLE_TEXT_X = 10.0
CONSOLE_INPUT_X_MONO = 26.0
CONSOLE_VERSION_OFFSET_X = 210.0
CONSOLE_VERSION_OFFSET_Y = 18.0
CONSOLE_BORDER_HEIGHT = 4.0
CONSOLE_BG_COLOR = (0.140625, 0.1875, 0.2890625)
CONSOLE_BORDER_COLOR = (0.21875, 0.265625, 0.3671875)
CONSOLE_PROMPT_MONO = ">"
CONSOLE_PROMPT_SMALL_FMT = ">%s"
CONSOLE_CARET_TEXT = "_"
SCRIPT_PAQ_NAMES = ("music.paq", "crimson.paq", "sfx.paq")

CommandHandler = Callable[[list[str]], None]


def game_build_path(base_dir: Path, name: str) -> Path:
    return base_dir / name


def _parse_float(value: str) -> float:
    try:
        return float(value)
    except ValueError:
        return 0.0


def _normalize_script_path(name: str) -> Path:
    raw = name.strip().strip("\"'")
    normalized = raw.replace("\\", "/")
    return Path(normalized)


def _resolve_script_path(console: "ConsoleState", target: Path) -> Path | None:
    if target.is_absolute():
        return target if target.is_file() else None
    for base in console.script_dirs:
        candidate = base / target
        if candidate.is_file():
            return candidate
    return None


def _primary_script_dirs(console: "ConsoleState") -> tuple[Path, ...]:
    dirs: list[Path] = [console.base_dir]
    if console.assets_dir is not None and console.assets_dir not in dirs:
        dirs.append(console.assets_dir)
    return tuple(dirs)


def _resolve_script_path_in(target: Path, roots: Iterable[Path]) -> Path | None:
    if target.is_absolute():
        return target if target.is_file() else None
    for base in roots:
        candidate = base / target
        if candidate.is_file():
            return candidate
    return None


def _iter_script_paq_paths(console: "ConsoleState") -> Iterable[Path]:
    roots: list[Path] = []
    if console.assets_dir is not None:
        roots.append(console.assets_dir)
    if console.base_dir not in roots:
        roots.append(console.base_dir)
    for root in roots:
        for name in SCRIPT_PAQ_NAMES:
            path = root / name
            if path.is_file():
                yield path


def _load_script_from_paq(console: "ConsoleState", target: Path) -> str | None:
    if target.is_absolute():
        return None
    normalized = target.as_posix().replace("\\", "/")
    normalized_lower = normalized.lower()
    for paq_path in _iter_script_paq_paths(console):
        try:
            for name, data in paq.iter_entries(paq_path):
                entry_name = name.replace("\\", "/")
                if entry_name == normalized or entry_name.lower() == normalized_lower:
                    return data.decode("utf-8", errors="ignore")
        except OSError:
            continue
        except Exception:
            continue
    return None


def _rgba(r: float, g: float, b: float, a: float) -> rl.Color:
    c = RGBA(r, g, b, a).clamped()
    return rl.Color(
        int(c.r * 255.0),
        int(c.g * 255.0),
        int(c.b * 255.0),
        int(c.a * 255.0),
    )


@dataclass(slots=True)
class ConsoleCvar:
    name: str
    value: str
    value_f: float

    @classmethod
    def from_value(cls, name: str, value: str) -> "ConsoleCvar":
        return cls(name=name, value=value, value_f=_parse_float(value))


@dataclass(slots=True)
class ConsoleLog:
    base_dir: Path
    lines: list[str] = field(default_factory=list)
    flushed_index: int = 0

    def log(self, message: str) -> None:
        self.lines.append(message)
        if len(self.lines) > MAX_CONSOLE_LINES:
            overflow = len(self.lines) - MAX_CONSOLE_LINES
            del self.lines[:overflow]
            self.flushed_index = max(0, self.flushed_index - overflow)

    def clear(self) -> None:
        self.lines.clear()
        self.flushed_index = 0

    def flush(self) -> None:
        if self.flushed_index >= len(self.lines):
            return
        path = game_build_path(self.base_dir, CONSOLE_LOG_NAME)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("a", encoding="utf-8") as handle:
            for line in self.lines[self.flushed_index :]:
                handle.write(line.rstrip() + "\n")
        self.flushed_index = len(self.lines)


@dataclass(slots=True)
class ConsoleState:
    base_dir: Path
    log: ConsoleLog
    assets_dir: Path | None = None
    script_dirs: tuple[Path, ...] = field(default_factory=tuple)
    commands: dict[str, CommandHandler] = field(default_factory=dict)
    cvars: dict[str, ConsoleCvar] = field(default_factory=dict)
    open_flag: bool = False
    input_enabled: bool = False
    input_ready: bool = False
    input_buffer: str = ""
    input_caret: int = 0
    history: list[str] = field(default_factory=list)
    history_index: int | None = None
    history_pending: str = ""
    scroll_offset: int = 0
    height_px: int = DEFAULT_CONSOLE_HEIGHT
    echo_enabled: bool = True
    quit_requested: bool = False
    prompt_string: str = "> %s"
    _mono_font: GrimMonoFont | None = field(default=None, init=False, repr=False)
    _mono_font_error: str | None = field(default=None, init=False, repr=False)
    _small_font: SmallFontData | None = field(default=None, init=False, repr=False)
    _small_font_error: str | None = field(default=None, init=False, repr=False)
    _slide_t: float = 1.0
    _offset_y: float = field(default=0.0, init=False)
    _blink_time: float = 0.0

    def register_command(self, name: str, handler: CommandHandler) -> None:
        self.commands[name] = handler

    def register_cvar(self, name: str, value: str) -> None:
        self.cvars[name] = ConsoleCvar.from_value(name, value)

    def add_script_dir(self, path: Path | None) -> None:
        if path is None:
            return
        if path in self.script_dirs:
            return
        self.script_dirs = (*self.script_dirs, path)

    def set_open(self, open_flag: bool) -> None:
        self.open_flag = open_flag
        self.input_enabled = open_flag
        self.input_ready = False
        self.history_index = None
        self._flush_input_queue()

    def toggle_open(self) -> None:
        self.set_open(not self.open_flag)

    def handle_hotkey(self) -> None:
        if rl.is_key_pressed(rl.KeyboardKey.KEY_GRAVE):
            self.toggle_open()

    def exec_line(self, line: str) -> None:
        tokens = self._tokenize_line(line)
        if not tokens:
            return
        name, args = tokens[0], tokens[1:]
        cvar = self.cvars.get(name)
        if cvar is not None:
            if args:
                value = " ".join(args)
                cvar.value = value
                cvar.value_f = _parse_float(value)
                self.log.log(f"\"{cvar.name}\" set to \"{cvar.value}\" ({cvar.value_f:.6f})")
            else:
                self.log.log(f"\"{cvar.name}\" is \"{cvar.value}\" ({cvar.value_f:.6f})")
            return
        handler = self.commands.get(name)
        if handler is not None:
            handler(args)
            return
        self.log.log(f"Unknown command \"{name}\"")

    def update(self, dt: float) -> None:
        frame_dt = min(dt, 0.1)
        self._blink_time += frame_dt
        self._update_slide(frame_dt)
        if not self.open_flag or not self.input_enabled:
            return
        ctrl_down = rl.is_key_down(rl.KeyboardKey.KEY_LEFT_CONTROL) or rl.is_key_down(rl.KeyboardKey.KEY_RIGHT_CONTROL)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_UP):
            if ctrl_down:
                self._scroll_lines(1)
            else:
                self._history_prev()
        if rl.is_key_pressed(rl.KeyboardKey.KEY_DOWN):
            if ctrl_down:
                self._scroll_lines(-1)
            else:
                self._history_next()
        if rl.is_key_pressed(rl.KeyboardKey.KEY_PAGE_UP):
            self._scroll_lines(2)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_PAGE_DOWN):
            self._scroll_lines(-2)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
            self.input_caret = max(0, self.input_caret - 1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
            self.input_caret = min(len(self.input_buffer), self.input_caret + 1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_HOME):
            self._scroll_lines(0x14)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_END):
            self.scroll_offset = 0
        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._autocomplete()
        if rl.is_key_pressed(rl.KeyboardKey.KEY_BACKSPACE):
            if self.input_caret > 0:
                self._exit_history_edit()
                self.input_buffer = (
                    self.input_buffer[: self.input_caret - 1] + self.input_buffer[self.input_caret :]
                )
                self.input_caret -= 1
        if rl.is_key_pressed(rl.KeyboardKey.KEY_DELETE):
            if self.input_caret < len(self.input_buffer):
                self._exit_history_edit()
                self.input_buffer = (
                    self.input_buffer[: self.input_caret] + self.input_buffer[self.input_caret + 1 :]
                )
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER):
            self._submit_input()
        self._poll_text_input()

    def draw(self) -> None:
        height = float(self.height_px)
        if height <= 0.0:
            return
        ratio = self._open_ratio(height)
        if ratio <= 0.0:
            return
        screen_w = float(rl.get_screen_width())
        offset_y = self._offset_y
        rl.draw_rectangle(
            0,
            int(offset_y),
            int(screen_w),
            int(height),
            _rgba(*CONSOLE_BG_COLOR, ratio),
        )
        border_y = int(offset_y + height - CONSOLE_BORDER_HEIGHT)
        rl.draw_rectangle(
            0,
            border_y,
            int(screen_w),
            int(CONSOLE_BORDER_HEIGHT),
            _rgba(*CONSOLE_BORDER_COLOR, ratio),
        )

        version_x = screen_w - CONSOLE_VERSION_OFFSET_X
        version_y = offset_y + height - CONSOLE_VERSION_OFFSET_Y
        self._draw_version_text(Vec2(version_x, version_y), _rgba(1.0, 1.0, 1.0, ratio * 0.3))

        visible, visible_count = self._visible_log_block(height)
        input_y = offset_y + (visible_count + 1) * CONSOLE_LINE_HEIGHT
        text_color = _rgba(1.0, 1.0, 1.0, ratio)
        use_mono = self._use_mono_font()
        if use_mono:
            self._draw_mono_text(CONSOLE_PROMPT_MONO, Vec2(CONSOLE_TEXT_X, input_y), text_color)
            self._draw_mono_text(self.input_buffer, Vec2(CONSOLE_INPUT_X_MONO, input_y), text_color)
        else:
            prompt = CONSOLE_PROMPT_SMALL_FMT.replace("%s", self.input_buffer)
            self._draw_small_text(prompt, Vec2(CONSOLE_TEXT_X, input_y), text_color)

        log_color = _rgba(0.6, 0.6, 0.7, ratio)
        y = offset_y + CONSOLE_LINE_HEIGHT
        for line in visible:
            if use_mono:
                self._draw_mono_text(line, Vec2(CONSOLE_TEXT_X, y), log_color)
            else:
                self._draw_small_text(line, Vec2(CONSOLE_TEXT_X, y), log_color)
            y += CONSOLE_LINE_HEIGHT

        caret_alpha = ratio * self._caret_blink_alpha()
        caret_color = _rgba(1.0, 1.0, 1.0, caret_alpha)
        caret_y = input_y + 2.0
        if use_mono:
            caret_x = CONSOLE_INPUT_X_MONO + float(self.input_caret) * 8.0
            self._draw_mono_text(CONSOLE_CARET_TEXT, Vec2(caret_x, caret_y), caret_color)
        else:
            caret_x = self._small_caret_x()
            self._draw_small_text(CONSOLE_CARET_TEXT, Vec2(caret_x, caret_y), caret_color)

    def close(self) -> None:
        if self._mono_font is not None:
            rl.unload_texture(self._mono_font.texture)
            self._mono_font = None
        if self._small_font is not None:
            rl.unload_texture(self._small_font.texture)
            self._small_font = None

    def _tokenize_line(self, line: str) -> list[str]:
        stripped = line.strip()
        if not stripped:
            return []
        if stripped.startswith("//"):
            return []
        return stripped.split()

    def _prompt_text(self) -> str:
        if "%s" in self.prompt_string:
            return self.prompt_string.replace("%s", self.input_buffer)
        return f"{self.prompt_string}{self.input_buffer}"

    def _history_prev(self) -> None:
        if not self.history:
            return
        if self.history_index is None:
            self.history_index = len(self.history) - 1
            self.history_pending = self.input_buffer
        elif self.history_index > 0:
            self.history_index -= 1
        self.input_buffer = self.history[self.history_index]
        self.input_caret = len(self.input_buffer)

    def _history_next(self) -> None:
        if self.history_index is None:
            return
        if self.history_index < len(self.history) - 1:
            self.history_index += 1
            self.input_buffer = self.history[self.history_index]
        else:
            self.history_index = None
            self.input_buffer = self.history_pending
        self.input_caret = len(self.input_buffer)

    def _exit_history_edit(self) -> None:
        if self.history_index is not None:
            self.history_index = None
            self.history_pending = self.input_buffer

    def _submit_input(self) -> None:
        line = self.input_buffer.strip()
        self.input_ready = True
        self.input_buffer = ""
        self.input_caret = 0
        self.history_index = None
        if not line:
            return
        if self.echo_enabled:
            if "%s" in self.prompt_string:
                self.log.log(self.prompt_string.replace("%s", line))
            else:
                self.log.log(f"{self.prompt_string}{line}")
        if not self.history or self.history[-1] != line:
            self.history.append(line)
        self.exec_line(line)
        self.input_ready = False
        self.scroll_offset = 0

    def _poll_text_input(self) -> None:
        while True:
            value = rl.get_char_pressed()
            if value == 0:
                break
            if value < 0x20 or value > 0xFF:
                continue
            if len(self.input_buffer) >= MAX_CONSOLE_INPUT:
                continue
            char = chr(value)
            self._exit_history_edit()
            self.input_buffer = (
                self.input_buffer[: self.input_caret] + char + self.input_buffer[self.input_caret :]
            )
            self.input_caret += 1

    def _autocomplete(self) -> None:
        if not self.input_buffer:
            return
        token_start = len(self.input_buffer) - len(self.input_buffer.lstrip())
        if token_start >= len(self.input_buffer):
            return
        token_end = self.input_buffer.find(" ", token_start)
        if token_end == -1:
            token_end = len(self.input_buffer)
        if self.input_caret > token_end:
            return
        prefix = self.input_buffer[token_start:self.input_caret]
        if not prefix:
            return
        match = self._autocomplete_name(prefix, self.cvars.keys())
        if match is None:
            match = self._autocomplete_name(prefix, self.commands.keys())
        if match is None:
            return
        self.input_buffer = self.input_buffer[:token_start] + match + self.input_buffer[token_end:]
        self.input_caret = token_start + len(match)

    def _autocomplete_name(self, prefix: str, names: Iterable[str]) -> str | None:
        for name in names:
            if name == prefix:
                return name
        for name in names:
            if name.startswith(prefix):
                return name
        return None

    def _scroll_lines(self, delta: int) -> None:
        max_offset = self._max_scroll_offset()
        if max_offset <= 0:
            self.scroll_offset = 0
            return
        self.scroll_offset = max(0, min(max_offset, self.scroll_offset + int(delta)))

    def _max_visible_lines(self, height: float | None = None) -> int:
        use_height = height if height is not None else float(self.height_px)
        if use_height <= 0.0:
            return 0
        return max(int(use_height // CONSOLE_LINE_HEIGHT) - 2, 0)

    def _max_scroll_offset(self) -> int:
        max_lines = self._max_visible_lines()
        log_count = len(self.log.lines)
        visible = min(log_count, max_lines)
        return max(0, log_count - visible)

    def _visible_log_block(self, height: float) -> tuple[list[str], int]:
        max_lines = self._max_visible_lines(height)
        log_count = len(self.log.lines)
        visible_count = min(log_count, max_lines)
        if visible_count <= 0:
            return [], 0
        max_offset = max(0, log_count - visible_count)
        if self.scroll_offset > max_offset:
            self.scroll_offset = max_offset
        start = max(0, log_count - visible_count - self.scroll_offset)
        end = min(log_count, start + visible_count)
        return self.log.lines[start:end], visible_count

    def _update_slide(self, dt: float) -> None:
        if self.open_flag:
            self._slide_t = max(0.0, self._slide_t - dt * CONSOLE_ANIM_SPEED)
        else:
            self._slide_t = min(1.0, self._slide_t + dt * CONSOLE_ANIM_SPEED)
        height = float(self.height_px)
        if height <= 0.0:
            self._offset_y = -height
            return
        eased = math.sin((1.0 - self._slide_t) * math.pi / 2.0)
        self._offset_y = eased * height - height

    def _open_ratio(self, height: float) -> float:
        if height <= 0.0:
            return 0.0
        return clamp((height + self._offset_y) / height, 0.0, 1.0)

    def _caret_blink_alpha(self) -> float:
        pulse = math.sin(self._blink_time * CONSOLE_BLINK_SPEED)
        value = max(0.2, abs(pulse) ** 2)
        return clamp(value, 0.0, 1.0)

    def _use_mono_font(self) -> bool:
        cvar = self.cvars.get("con_monoFont")
        if cvar is None:
            return False
        return bool(cvar.value_f)

    def _ensure_mono_font(self) -> GrimMonoFont | None:
        if self._mono_font is not None:
            return self._mono_font
        if self._mono_font_error is not None:
            return None
        if self.assets_dir is None:
            self._mono_font_error = "missing assets dir"
            return None
        missing_assets: list[str] = []
        try:
            self._mono_font = load_grim_mono_font(self.assets_dir, missing_assets)
        except FileNotFoundError as exc:
            self._mono_font_error = str(exc)
            self._mono_font = None
        return self._mono_font

    def _ensure_small_font(self) -> SmallFontData | None:
        if self._small_font is not None:
            return self._small_font
        if self._small_font_error is not None:
            return None
        if self.assets_dir is None:
            self._small_font_error = "missing assets dir"
            return None
        missing_assets: list[str] = []
        try:
            self._small_font = load_small_font(self.assets_dir, missing_assets)
        except FileNotFoundError as exc:
            self._small_font_error = str(exc)
            self._small_font = None
        return self._small_font

    def _draw_mono_text(self, text: str, pos: Vec2, color: rl.Color) -> None:
        font = self._ensure_mono_font()
        if font is None:
            rl.draw_text(text, int(pos.x), int(pos.y), int(16 * CONSOLE_MONO_SCALE), color)
            return
        advance = font.advance * CONSOLE_MONO_SCALE
        draw_grim_mono_text(font, text, Vec2(pos.x - advance, pos.y), CONSOLE_MONO_SCALE, color)

    def _draw_small_text(self, text: str, pos: Vec2, color: rl.Color) -> None:
        font = self._ensure_small_font()
        if font is None:
            rl.draw_text(text, int(pos.x), int(pos.y), int(16 * CONSOLE_SMALL_SCALE), color)
            return
        draw_small_text(font, text, pos, CONSOLE_SMALL_SCALE, color)

    def _draw_version_text(self, pos: Vec2, color: rl.Color) -> None:
        font = self._ensure_small_font()
        if font is None:
            self._draw_mono_text(CONSOLE_VERSION_TEXT, pos, color)
            return
        draw_small_text(font, CONSOLE_VERSION_TEXT, pos, CONSOLE_SMALL_SCALE, color)

    def _small_caret_x(self) -> float:
        font = self._ensure_small_font()
        if font is None:
            return CONSOLE_TEXT_X + 16.0 + float(self.input_caret) * 8.0
        prompt_w = measure_small_text_width(font, CONSOLE_PROMPT_SMALL_FMT.replace("%s", ""), CONSOLE_SMALL_SCALE)
        input_w = measure_small_text_width(font, self.input_buffer[: self.input_caret], CONSOLE_SMALL_SCALE)
        return CONSOLE_TEXT_X + prompt_w + input_w

    def _flush_input_queue(self) -> None:
        while rl.get_char_pressed():
            pass
        while rl.get_key_pressed():
            pass


def create_console(base_dir: Path, assets_dir: Path | None = None) -> ConsoleState:
    script_dirs: tuple[Path, ...] = (base_dir,)
    if assets_dir is not None and assets_dir != base_dir:
        script_dirs = (*script_dirs, assets_dir)
    console = ConsoleState(
        base_dir=base_dir,
        log=ConsoleLog(base_dir=base_dir),
        assets_dir=assets_dir,
        script_dirs=script_dirs,
    )
    console.register_cvar("version", CONSOLE_VERSION_TEXT)
    console.register_cvar("con_monoFont", "1")
    if console.open_flag:
        console._slide_t = 0.0
        console._offset_y = 0.0
    else:
        console._slide_t = 1.0
        console._offset_y = -float(console.height_px)
    register_core_commands(console)
    return console


def _make_noop_command(console: ConsoleState, name: str) -> CommandHandler:
    def _handler(args: list[str]) -> None:
        console.log.log(f"command {name} called with {len(args)} args")

    return _handler


def register_boot_commands(
    console: ConsoleState, handlers: dict[str, CommandHandler] | None = None
) -> None:
    resolved = handlers or {}
    commands = (
        "setGammaRamp",
        "snd_addGameTune",
        "generateterrain",
        "telltimesurvived",
        "setresourcepaq",
        "loadtexture",
        "openurl",
        "sndfreqadjustment",
    )
    for name in commands:
        handler = resolved.get(name)
        if handler is None:
            handler = _make_noop_command(console, name)
        console.register_command(name, handler)


def register_core_cvars(console: ConsoleState, width: int, height: int) -> None:
    # Defaults match `register_core_cvars` (0x00402c00) in `crimsonland.exe`.
    console.register_cvar("cv_silentloads", "1")
    console.register_cvar("cv_terrainFilter", "1")
    console.register_cvar("cv_bodiesFade", "1")
    console.register_cvar("cv_uiTransparency", "1")
    console.register_cvar("cv_uiPointFilterPanels", "0")
    console.register_cvar("cv_enableMousePointAndClickMovement", "0")
    console.register_cvar("cv_verbose", "0")
    console.register_cvar("cv_terrainBodiesTransparency", "0")
    console.register_cvar("cv_uiSmallIndicators", "0")
    console.register_cvar("cv_aimEnhancementFade", "0.7")
    console.register_cvar("cv_friendlyFire", "0")
    console.register_cvar("cv_showFPS", "0")
    console.register_cvar("cv_padAimDistMul", "96")
    console.register_cvar("v_width", str(width))
    console.register_cvar("v_height", str(height))


def register_core_commands(console: ConsoleState) -> None:
    def cmdlist(_args: list[str]) -> None:
        for name in console.commands.keys():
            console.log.log(name)
        console.log.log(f"{len(console.commands)} commands")

    def vars_cmd(_args: list[str]) -> None:
        for name in console.cvars.keys():
            console.log.log(name)
        console.log.log(f"{len(console.cvars)} variables")

    def cmd_set(args: list[str]) -> None:
        if len(args) < 2:
            console.log.log("Usage: set <var> <value>")
            return
        name = args[0]
        value = " ".join(args[1:])
        console.register_cvar(name, value)
        console.log.log(f"'{name}' set to '{value}'")

    def cmd_echo(args: list[str]) -> None:
        if not args:
            console.log.log(f"echo is {'on' if console.echo_enabled else 'off'}")
            return
        mode = args[0].lower()
        if mode in {"on", "off"}:
            console.echo_enabled = mode == "on"
            console.log.log(f"echo {mode}")
            return
        console.log.log(" ".join(args))

    def cmd_quit(_args: list[str]) -> None:
        console.quit_requested = True

    def cmd_clear(_args: list[str]) -> None:
        console.log.clear()
        console.scroll_offset = 0

    def cmd_extend(_args: list[str]) -> None:
        console.height_px = EXTENDED_CONSOLE_HEIGHT

    def cmd_minimize(_args: list[str]) -> None:
        console.height_px = DEFAULT_CONSOLE_HEIGHT

    def cmd_exec(args: list[str]) -> None:
        if not args:
            console.log.log("exec <script>")
            return
        target = _normalize_script_path(args[0])
        try:
            script_text: str | None = None
            primary_dirs = _primary_script_dirs(console)
            path = _resolve_script_path_in(target, primary_dirs)
            if path is None:
                script_text = _load_script_from_paq(console, target)
            if path is None and script_text is None:
                fallback_dirs = [path for path in console.script_dirs if path not in primary_dirs]
                path = _resolve_script_path_in(target, fallback_dirs)
            if path is None and script_text is None:
                console.log.log(f"Cannot open file '{args[0]}'")
                return
            console.log.log(f"Executing '{args[0]}'")
            if script_text is None:
                assert path is not None
                script_text = path.read_text(encoding="utf-8", errors="ignore")
            for raw_line in script_text.splitlines():
                line = raw_line.strip()
                if line:
                    console.exec_line(line)
        except OSError:
            console.log.log(f"Cannot open file '{args[0]}'")

    console.register_command("cmdlist", cmdlist)
    console.register_command("vars", vars_cmd)
    console.register_command("set", cmd_set)
    console.register_command("echo", cmd_echo)
    console.register_command("quit", cmd_quit)
    console.register_command("clear", cmd_clear)
    console.register_command("extendconsole", cmd_extend)
    console.register_command("minimizeconsole", cmd_minimize)
    console.register_command("exec", cmd_exec)



from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class FixedStepClock:
    tick_rate: int = 60
    accum: float = 0.0

    def __post_init__(self) -> None:
        tick_rate = int(self.tick_rate)
        if tick_rate <= 0:
            raise ValueError(f"tick_rate must be positive, got {tick_rate}")
        self.tick_rate = tick_rate
        self.accum = float(self.accum)

    @property
    def dt_tick(self) -> float:
        return 1.0 / float(self.tick_rate)

    def reset(self) -> None:
        self.accum = 0.0

    def advance(self, dt: float, *, max_dt: float = 0.1) -> int:
        dt = float(dt)
        if dt <= 0.0:
            return 0
        if dt > float(max_dt):
            dt = float(max_dt)

        self.accum += dt
        dt_tick = float(self.dt_tick)
        if not (dt_tick > 0.0):
            return 0

        ticks = int((self.accum + 1e-9) / dt_tick)
        if ticks <= 0:
            return 0

        self.accum -= dt_tick * float(ticks)
        if self.accum < 0.0:
            self.accum = 0.0
        return int(ticks)




from __future__ import annotations

import pyray as rl

from .shadow import UI_SHADOW_OFFSET, draw_ui_quad_shadow


# Classic menu panel is rendered from the *inset* inner region of ui_menuPanel:
#   - X inset: 1px on each side (uv 1/512 .. 511/512) => 510px wide
#   - Y inset: 1px on each side (uv 1/256 .. 255/256) => 254px tall
#
# When a panel is taller than the base height, the original stretches it using a
# 3-slice: [top][mid][bottom]. The source slice boundaries are at y=130 and y=150
# in the texture (see grim UVs in ui_render_trace).
MENU_PANEL_INSET = 1.0
MENU_PANEL_SRC_SLICE_Y1 = 130.0
MENU_PANEL_SRC_SLICE_Y2 = 150.0

# Destination slice heights observed in the original at scale=1.0 (1024x768).
MENU_PANEL_DST_TOP_H = 138.0
MENU_PANEL_DST_BOTTOM_H = 116.0


def draw_classic_menu_panel(
    texture: rl.Texture,
    *,
    dst: rl.Rectangle,
    tint: rl.Color = rl.WHITE,
    shadow: bool = False,
    flip_x: bool = False,
) -> None:
    """
    Draw a classic menu panel (ui_menuPanel) with the same slicing behavior as the original.

    - Uses inset source rect (1px border skipped) to match the vertex/UV inset.
    - Uses 3-slice only when dst is taller than (top + bottom); otherwise draws a single quad.
    """

    tex_w = float(texture.width)
    tex_h = float(texture.height)
    if tex_w <= 0.0 or tex_h <= 0.0:
        return

    inset = MENU_PANEL_INSET
    src_x = inset
    src_y = inset
    src_w = max(0.0, tex_w - inset * 2.0)
    src_h = max(0.0, tex_h - inset * 2.0)

    # Scale slice heights with the panel width (menu panel uses the same scale factor).
    # dst.width is already in our "inset" width space (510 at scale=1.0).
    scale = (float(dst.width) / 510.0) if float(dst.width) != 0.0 else 1.0
    top_h = MENU_PANEL_DST_TOP_H * scale
    bottom_h = MENU_PANEL_DST_BOTTOM_H * scale
    mid_h = float(dst.height) - top_h - bottom_h

    origin = rl.Vector2(0.0, 0.0)

    def _src(rect: rl.Rectangle) -> rl.Rectangle:
        if not flip_x:
            return rect
        return rl.Rectangle(rect.x + rect.width, rect.y, -rect.width, rect.height)

    if mid_h <= 0.0:
        src = _src(rl.Rectangle(src_x, src_y, src_w, src_h))
        if shadow:
            draw_ui_quad_shadow(
                texture=texture,
                src=src,
                dst=rl.Rectangle(
                    float(dst.x + UI_SHADOW_OFFSET),
                    float(dst.y + UI_SHADOW_OFFSET),
                    float(dst.width),
                    float(dst.height),
                ),
                origin=origin,
                rotation_deg=0.0,
            )
        rl.draw_texture_pro(texture, src, dst, origin, 0.0, tint)
        return

    # Source slice rects (in texture pixels, with 1px inset).
    src_top = _src(rl.Rectangle(src_x, src_y, src_w, max(0.0, MENU_PANEL_SRC_SLICE_Y1 - inset)))
    src_mid = _src(
        rl.Rectangle(src_x, MENU_PANEL_SRC_SLICE_Y1, src_w, max(0.0, MENU_PANEL_SRC_SLICE_Y2 - MENU_PANEL_SRC_SLICE_Y1))
    )
    src_bot = _src(rl.Rectangle(src_x, MENU_PANEL_SRC_SLICE_Y2, src_w, max(0.0, (tex_h - inset) - MENU_PANEL_SRC_SLICE_Y2)))

    # Destination slices.
    dst_top = rl.Rectangle(dst.x, dst.y, float(dst.width), float(top_h))
    dst_mid = rl.Rectangle(dst.x, dst.y + float(top_h), float(dst.width), float(mid_h))
    dst_bot = rl.Rectangle(dst.x, dst.y + float(top_h) + float(mid_h), float(dst.width), float(bottom_h))

    if shadow:
        draw_ui_quad_shadow(
            texture=texture,
            src=src_top,
            dst=rl.Rectangle(
                float(dst_top.x + UI_SHADOW_OFFSET),
                float(dst_top.y + UI_SHADOW_OFFSET),
                float(dst_top.width),
                float(dst_top.height),
            ),
            origin=origin,
            rotation_deg=0.0,
        )
        draw_ui_quad_shadow(
            texture=texture,
            src=src_mid,
            dst=rl.Rectangle(
                float(dst_mid.x + UI_SHADOW_OFFSET),
                float(dst_mid.y + UI_SHADOW_OFFSET),
                float(dst_mid.width),
                float(dst_mid.height),
            ),
            origin=origin,
            rotation_deg=0.0,
        )
        draw_ui_quad_shadow(
            texture=texture,
            src=src_bot,
            dst=rl.Rectangle(
                float(dst_bot.x + UI_SHADOW_OFFSET),
                float(dst_bot.y + UI_SHADOW_OFFSET),
                float(dst_bot.width),
                float(dst_bot.height),
            ),
            origin=origin,
            rotation_deg=0.0,
        )

    rl.draw_texture_pro(texture, src_top, dst_top, origin, 0.0, tint)
    rl.draw_texture_pro(texture, src_mid, dst_mid, origin, 0.0, tint)
    rl.draw_texture_pro(texture, src_bot, dst_bot, origin, 0.0, tint)



from __future__ import annotations

from dataclasses import dataclass, field
import math
from typing import Callable, Sequence

from grim.geom import Vec2

from ..creatures.runtime import CreatureDeath
from ..creatures.spawn import CreatureTypeId
from ..effects import FxQueue
from ..features.bonuses.freeze import freeze_bonus_active
from ..features.presentation import ProjectileDecalCtx, run_projectile_decal_hooks
from ..game_modes import GameMode
from ..gameplay import BonusPickupEvent, GameplayState, PlayerState, perk_active
from ..perks import PerkId
from ..projectiles import ProjectileHit
from ..weapon_sfx import resolve_weapon_sfx_ref
from ..weapons import WEAPON_BY_ID, WeaponId
from .world_defs import BEAM_TYPES

_MAX_HIT_SFX_PER_FRAME = 4
_MAX_DEATH_SFX_PER_FRAME = 5

_BULLET_HIT_SFX = (
    "sfx_bullet_hit_01",
    "sfx_bullet_hit_02",
    "sfx_bullet_hit_03",
    "sfx_bullet_hit_04",
    "sfx_bullet_hit_05",
    "sfx_bullet_hit_06",
)

_CREATURE_DEATH_SFX: dict[CreatureTypeId, tuple[str, ...]] = {
    CreatureTypeId.ZOMBIE: (
        "sfx_zombie_die_01",
        "sfx_zombie_die_02",
        "sfx_zombie_die_03",
        "sfx_zombie_die_04",
    ),
    CreatureTypeId.LIZARD: (
        "sfx_lizard_die_01",
        "sfx_lizard_die_02",
        "sfx_lizard_die_03",
        "sfx_lizard_die_04",
    ),
    CreatureTypeId.ALIEN: (
        "sfx_alien_die_01",
        "sfx_alien_die_02",
        "sfx_alien_die_03",
        "sfx_alien_die_04",
    ),
    CreatureTypeId.SPIDER_SP1: (
        "sfx_spider_die_01",
        "sfx_spider_die_02",
        "sfx_spider_die_03",
        "sfx_spider_die_04",
    ),
    CreatureTypeId.SPIDER_SP2: (
        "sfx_spider_die_01",
        "sfx_spider_die_02",
        "sfx_spider_die_03",
        "sfx_spider_die_04",
    ),
    CreatureTypeId.TROOPER: (
        "sfx_trooper_die_01",
        "sfx_trooper_die_02",
        "sfx_trooper_die_03",
        "sfx_trooper_die_04",
    ),
}


@dataclass(slots=True)
class PresentationStepCommands:
    trigger_game_tune: bool = False
    sfx_keys: list[str] = field(default_factory=list)


def plan_player_audio_sfx(
    player: object,
    *,
    prev_shot_seq: int,
    prev_reload_active: bool,
    prev_reload_timer: float,
) -> list[str]:
    keys: list[str] = []

    weapon = WEAPON_BY_ID.get(int(getattr(player, "weapon_id", 0)))
    if weapon is None:
        return keys

    if int(getattr(player, "shot_seq", 0)) > int(prev_shot_seq):
        if float(getattr(player, "fire_bullets_timer", 0.0)) > 0.0:
            fire_bullets = WEAPON_BY_ID.get(int(WeaponId.FIRE_BULLETS))
            plasma_minigun = WEAPON_BY_ID.get(int(WeaponId.PLASMA_MINIGUN))
            if fire_bullets is not None:
                key = resolve_weapon_sfx_ref(fire_bullets.fire_sound)
                if key is not None:
                    keys.append(key)
            if plasma_minigun is not None:
                key = resolve_weapon_sfx_ref(plasma_minigun.fire_sound)
                if key is not None:
                    keys.append(key)
        else:
            key = resolve_weapon_sfx_ref(weapon.fire_sound)
            if key is not None:
                keys.append(key)

    reload_active = bool(getattr(player, "reload_active", False))
    reload_timer = float(getattr(player, "reload_timer", 0.0))
    reload_started = (not prev_reload_active and reload_active) or (reload_timer > prev_reload_timer + 1e-6)
    if reload_started:
        key = resolve_weapon_sfx_ref(weapon.reload_sound)
        if key is not None:
            keys.append(key)

    return keys


def _rand_choice(rand: Callable[[], int], options: tuple[str, ...]) -> str | None:
    if not options:
        return None
    idx = int(rand()) % len(options)
    return options[idx]


def _hit_sfx_for_type(
    type_id: int,
    *,
    beam_types: frozenset[int],
    rand: Callable[[], int],
) -> str | None:
    _ = beam_types
    weapon = WEAPON_BY_ID.get(int(type_id))
    ammo_class = weapon.ammo_class if weapon is not None else None
    if ammo_class == 4:
        return "sfx_shock_hit_01"
    return _rand_choice(rand, _BULLET_HIT_SFX)


def plan_hit_sfx_keys(
    hits: list[ProjectileHit],
    *,
    game_mode: int,
    demo_mode_active: bool,
    game_tune_started: bool,
    rand: Callable[[], int],
    beam_types: frozenset[int] = BEAM_TYPES,
) -> tuple[bool, list[str]]:
    if not hits:
        return False, []

    trigger_game_tune = False
    if (not demo_mode_active) and int(game_mode) != int(GameMode.RUSH) and (not game_tune_started):
        trigger_game_tune = True

    end = min(len(hits), _MAX_HIT_SFX_PER_FRAME)
    keys: list[str] = []
    for idx in range(0, end):
        type_id = int(hits[idx].type_id)
        key = _hit_sfx_for_type(type_id, beam_types=beam_types, rand=rand)
        if key is not None:
            keys.append(key)
    return trigger_game_tune, keys


def plan_death_sfx_keys(
    deaths: Sequence[CreatureDeath] | tuple[object, ...],
    *,
    rand: Callable[[], int],
) -> list[str]:
    keys: list[str] = []
    if not deaths:
        return keys

    for idx in range(min(len(deaths), _MAX_DEATH_SFX_PER_FRAME)):
        death = deaths[idx]
        type_id = getattr(death, "type_id", None)
        if type_id is None:
            continue
        try:
            creature_type = CreatureTypeId(int(type_id))
        except ValueError:
            continue
        options = _CREATURE_DEATH_SFX.get(creature_type)
        if options:
            key = _rand_choice(rand, options)
            if key is not None:
                keys.append(key)
    return keys


@dataclass(frozen=True, slots=True)
class ProjectileDecalPostCtx:
    hit: ProjectileHit
    base_angle: float
    type_id: int
    freeze_active: bool


def queue_projectile_decals(
    *,
    state: GameplayState,
    players: Sequence[PlayerState],
    fx_queue: FxQueue,
    hits: list[ProjectileHit],
    rand: Callable[[], int],
    detail_preset: int,
    fx_toggle: int,
) -> None:
    for hit in hits:
        post_ctx = queue_projectile_decals_pre_hit(
            state=state,
            players=players,
            fx_queue=fx_queue,
            hit=hit,
            rand=rand,
            detail_preset=detail_preset,
            fx_toggle=fx_toggle,
        )
        queue_projectile_decals_post_hit(
            fx_queue=fx_queue,
            post_ctx=post_ctx,
            rand=rand,
        )


def queue_projectile_decals_pre_hit(
    *,
    state: GameplayState,
    players: Sequence[PlayerState],
    fx_queue: FxQueue,
    hit: ProjectileHit,
    rand: Callable[[], int],
    detail_preset: int,
    fx_toggle: int,
) -> ProjectileDecalPostCtx:
    freeze_active = freeze_bonus_active(state=state)
    bloody = bool(players) and perk_active(players[0], PerkId.BLOODY_MESS_QUICK_LEARNER)

    type_id = int(hit.type_id)

    base_angle = (hit.hit - hit.origin).to_angle()

    # Native `projectile_update` spawns blood splatter before terrain decals.
    if bloody:
        for _ in range(8):
            spread = float((int(rand()) & 0x1F) - 0x10) * 0.0625
            state.effects.spawn_blood_splatter(
                pos=hit.hit,
                angle=base_angle + spread,
                age=0.0,
                rand=rand,
                detail_preset=detail_preset,
                fx_toggle=fx_toggle,
            )
        state.effects.spawn_blood_splatter(
            pos=hit.hit,
            angle=base_angle + math.pi,
            age=0.0,
            rand=rand,
            detail_preset=detail_preset,
            fx_toggle=fx_toggle,
        )

        lo = -30
        hi = 30
        while lo > -60:
            span = hi - lo
            for _ in range(2):
                dx = float(int(rand()) % span + lo)
                dy = float(int(rand()) % span + lo)
                fx_queue.add_random(
                    pos=hit.target + Vec2(dx, dy),
                    rand=rand,
                )
            lo -= 10
            hi += 10
    elif not freeze_active:
        for _ in range(2):
            state.effects.spawn_blood_splatter(
                pos=hit.hit,
                angle=base_angle,
                age=0.0,
                rand=rand,
                detail_preset=detail_preset,
                fx_toggle=fx_toggle,
            )
            if (int(rand()) & 7) == 2:
                state.effects.spawn_blood_splatter(
                    pos=hit.hit,
                    angle=base_angle + math.pi,
                    age=0.0,
                    rand=rand,
                    detail_preset=detail_preset,
                    fx_toggle=fx_toggle,
                )

    return ProjectileDecalPostCtx(
        hit=hit,
        base_angle=float(base_angle),
        type_id=int(type_id),
        freeze_active=bool(freeze_active),
    )


def queue_projectile_decals_post_hit(
    *,
    fx_queue: FxQueue,
    post_ctx: ProjectileDecalPostCtx,
    rand: Callable[[], int],
) -> None:
    hit = post_ctx.hit
    base_angle = float(post_ctx.base_angle)

    # Native consumes one extra `crt_rand()` per creature hit before the
    # post-hit terrain decal burst branch.
    rand()

    hook_handled = run_projectile_decal_hooks(
        ProjectileDecalCtx(
            hit=hit,
            base_angle=float(base_angle),
            fx_queue=fx_queue,
            rand=rand,
        )
    )

    if bool(hook_handled) or bool(post_ctx.freeze_active):
        return

    for _ in range(3):
        spread = float(int(rand()) % 0x14 - 10) * 0.1
        angle = base_angle + spread
        direction = Vec2.from_angle(angle) * 20.0
        fx_queue.add_random(pos=hit.target, rand=rand)
        fx_queue.add_random(
            pos=hit.target + direction * 1.5,
            rand=rand,
        )
        fx_queue.add_random(
            pos=hit.target + direction * 2.0,
            rand=rand,
        )
        fx_queue.add_random(
            pos=hit.target + direction * 2.5,
            rand=rand,
        )


def apply_world_presentation_step(
    *,
    state: GameplayState,
    players: Sequence[PlayerState],
    fx_queue: FxQueue,
    hits: list[ProjectileHit],
    deaths: tuple[CreatureDeath, ...],
    pickups: list[BonusPickupEvent],
    event_sfx: list[str],
    prev_audio: Sequence[tuple[int, bool, float]],
    prev_perk_pending: int,
    game_mode: int,
    demo_mode_active: bool,
    perk_progression_enabled: bool,
    rand: Callable[[], int],
    rand_for: Callable[[str], Callable[[], int]] | None = None,
    detail_preset: int,
    fx_toggle: int,
    game_tune_started: bool,
    trigger_game_tune: bool | None = None,
    hit_sfx: Sequence[str] | None = None,
    death_sfx_preplanned: bool = False,
) -> PresentationStepCommands:
    commands = PresentationStepCommands()
    if rand_for is None:
        def rand_for(_label: str) -> Callable[[], int]:
            return rand
    if perk_progression_enabled and int(state.perk_selection.pending_count) > int(prev_perk_pending):
        commands.sfx_keys.append("sfx_ui_levelup")
    if trigger_game_tune is None and hit_sfx is None:
        if hits:
            queue_projectile_decals(
                state=state,
                players=players,
                fx_queue=fx_queue,
                hits=hits,
                rand=rand_for("projectile_decals"),
                detail_preset=int(detail_preset),
                fx_toggle=int(fx_toggle),
            )
            if freeze_bonus_active(state=state):
                if (not bool(demo_mode_active)) and int(game_mode) != int(GameMode.RUSH) and (not bool(game_tune_started)):
                    commands.trigger_game_tune = True
            else:
                commands.trigger_game_tune, planned_hit_sfx = plan_hit_sfx_keys(
                    hits,
                    game_mode=int(game_mode),
                    demo_mode_active=bool(demo_mode_active),
                    game_tune_started=bool(game_tune_started),
                    rand=rand_for("hit_sfx"),
                )
                commands.sfx_keys.extend(planned_hit_sfx)
    else:
        if trigger_game_tune is not None:
            commands.trigger_game_tune = bool(trigger_game_tune)
        if hit_sfx is not None:
            commands.sfx_keys.extend(str(key) for key in hit_sfx)
    for idx, player in enumerate(players):
        if idx >= len(prev_audio):
            continue
        prev_shot_seq, prev_reload_active, prev_reload_timer = prev_audio[idx]
        commands.sfx_keys.extend(
            plan_player_audio_sfx(
                player,
                prev_shot_seq=int(prev_shot_seq),
                prev_reload_active=bool(prev_reload_active),
                prev_reload_timer=float(prev_reload_timer),
            )
        )
    if deaths and not death_sfx_preplanned:
        commands.sfx_keys.extend(plan_death_sfx_keys(deaths, rand=rand_for("death_sfx")))
    if pickups:
        commands.sfx_keys.extend("sfx_ui_bonus" for _ in pickups)
    commands.sfx_keys.extend(str(key) for key in event_sfx[:4])
    return commands



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
import datetime as dt
import hashlib
import random

import pyray as rl

from grim.assets import PaqTextureCache
from grim.audio import AudioState
from grim.console import ConsoleState
from grim.config import CrimsonConfig
from grim.geom import Vec2
from grim.view import ViewContext

from ..game_modes import GameMode
from ..gameplay import weapon_assign_player
from ..ui.cursor import draw_aim_cursor, draw_menu_cursor
from ..ui.hud import draw_hud_overlay, hud_flags_for_game_mode
from ..ui.perk_menu import load_perk_menu_assets
from ..replay import ReplayHeader, ReplayRecorder, ReplayStatusSnapshot, dump_replay
from ..replay.types import WEAPON_USAGE_COUNT
from ..replay.checkpoints import (
    FORMAT_VERSION as CHECKPOINTS_FORMAT_VERSION,
    ReplayCheckpoint,
    ReplayCheckpoints,
    build_checkpoint,
    default_checkpoints_path,
    dump_checkpoints_file,
    resolve_checkpoint_sample_rate,
)
from ..sim.clock import FixedStepClock
from ..sim.sessions import RushDeterministicSession
from ..weapons import WeaponId
from .base_gameplay_mode import BaseGameplayMode
from .components.highscore_record_builder import build_highscore_record_for_game_over

WORLD_SIZE = 1024.0
RUSH_WEAPON_ID = WeaponId.ASSAULT_RIFLE

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)


@dataclass(slots=True)
class _RushState:
    elapsed_ms: float = 0.0
    spawn_cooldown_ms: float = 0.0


class RushMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=int(GameMode.RUSH),
            demo_mode_active=False,
            difficulty_level=0,
            hardcore=False,
            texture_cache=texture_cache,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._rush = _RushState()

        self._ui_assets = None
        self._sim_clock = FixedStepClock(tick_rate=60)
        self._replay_recorder: ReplayRecorder | None = None
        self._replay_checkpoints: list[ReplayCheckpoint] = []
        self._replay_checkpoints_sample_rate: int = 60
        self._replay_checkpoints_last_tick: int | None = None
        self._sim_session: RushDeterministicSession | None = None

    def _record_replay_checkpoint(
        self,
        tick_index: int,
        *,
        force: bool = False,
        rng_marks: dict[str, int] | None = None,
        deaths: Sequence[object] | None = None,
        events: object | None = None,
        command_hash: str | None = None,
    ) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        if tick_index < 0:
            return
        if not force and (tick_index % int(self._replay_checkpoints_sample_rate or 1)) != 0:
            return
        if self._replay_checkpoints_last_tick == int(tick_index):
            return
        self._replay_checkpoints.append(
            build_checkpoint(
                tick_index=int(tick_index),
                world=self._world.world_state,
                elapsed_ms=float(self._rush.elapsed_ms),
                rng_marks=rng_marks,
                deaths=deaths,
                events=events,
                command_hash=command_hash,
            )
        )
        self._replay_checkpoints_last_tick = int(tick_index)

    def _enforce_rush_loadout(self) -> None:
        for player in self._world.players:
            if int(player.weapon_id) != RUSH_WEAPON_ID:
                weapon_assign_player(player, RUSH_WEAPON_ID)
            # `rush_mode_update` forces weapon+ammo every frame; keep ammo topped up.
            player.ammo = float(max(0, int(player.clip_size)))

    def open(self) -> None:
        super().open()
        self._ui_assets = load_perk_menu_assets(self._assets_root)
        if self._ui_assets.missing:
            self._missing_assets.extend(self._ui_assets.missing)
        self._rush = _RushState()
        self._sim_clock.reset()
        self._sim_session = RushDeterministicSession(
            world=self._world.world_state,
            world_size=float(self._world.world_size),
            damage_scale_by_type=self._world._damage_scale_by_type,
            fx_queue=self._world.fx_queue,
            fx_queue_rotated=self._world.fx_queue_rotated,
            detail_preset=5,
            fx_toggle=0,
            game_tune_started=bool(self._world._game_tune_started),
            clear_fx_queues_each_tick=False,
            enforce_loadout=self._enforce_rush_loadout,
        )
        self._enforce_rush_loadout()
        status = self._state.status
        weapon_usage_counts: tuple[int, ...] = ()
        if status is not None:
            try:
                raw_counts = status.data.get("weapon_usage_counts")
                if isinstance(raw_counts, list):
                    weapon_usage_counts = tuple(int(value) & 0xFFFFFFFF for value in raw_counts[:WEAPON_USAGE_COUNT])
            except Exception:
                weapon_usage_counts = ()
        if len(weapon_usage_counts) != WEAPON_USAGE_COUNT:
            weapon_usage_counts = tuple(weapon_usage_counts) + (0,) * max(
                0, WEAPON_USAGE_COUNT - len(weapon_usage_counts)
            )
            weapon_usage_counts = weapon_usage_counts[:WEAPON_USAGE_COUNT]
        status_snapshot = ReplayStatusSnapshot(
            quest_unlock_index=int(getattr(status, "quest_unlock_index", 0) or 0) if status is not None else 0,
            quest_unlock_index_full=int(getattr(status, "quest_unlock_index_full", 0) or 0)
            if status is not None
            else 0,
            weapon_usage_counts=weapon_usage_counts,
        )
        self._replay_recorder = ReplayRecorder(
            ReplayHeader(
                game_mode_id=int(GameMode.RUSH),
                seed=int(self._state.rng.state),
                tick_rate=int(self._sim_clock.tick_rate),
                difficulty_level=int(self._world.difficulty_level),
                hardcore=bool(self._world.hardcore),
                preserve_bugs=bool(self._world.preserve_bugs),
                world_size=float(self._world.world_size),
                player_count=len(self._world.players),
                status=status_snapshot,
            )
        )
        tick_rate = int(self._replay_recorder.header.tick_rate)
        self._replay_checkpoints_sample_rate = resolve_checkpoint_sample_rate(tick_rate)
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

    def close(self) -> None:
        if self._ui_assets is not None:
            self._ui_assets = None
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        self._sim_session = None
        super().close()

    def _handle_input(self) -> None:
        if self._game_over_active:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
                self._action = "back_to_menu"
                self.close_requested = True
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _player_name_default(self) -> str:
        config = self._config
        if config is None:
            return ""
        raw = config.data.get("player_name")
        if isinstance(raw, (bytes, bytearray)):
            return bytes(raw).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")
        if isinstance(raw, str):
            return raw
        return ""

    def _enter_game_over(self) -> None:
        if self._game_over_active:
            return

        game_mode_id = (
            int(self._config.data.get("game_mode", int(GameMode.RUSH)))
            if self._config is not None
            else int(GameMode.RUSH)
        )
        record = build_highscore_record_for_game_over(
            state=self._state,
            player=self._player,
            survival_elapsed_ms=int(self._rush.elapsed_ms),
            creature_kill_count=int(self._creatures.kill_count),
            game_mode_id=game_mode_id,
        )

        self._game_over_record = record
        self._game_over_ui.open()
        self._game_over_active = True
        self._save_replay()

    def _save_replay(self) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        self._record_replay_checkpoint(max(0, recorder.tick_index - 1), force=True)
        replay = recorder.finish()
        data = dump_replay(replay)
        digest = hashlib.sha256(data).hexdigest()
        stamp = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
        replay_dir = self._base_dir / "replays"
        replay_dir.mkdir(parents=True, exist_ok=True)
        kills = int(self._creatures.kill_count)
        base_name = f"rush_{stamp}_kills{kills}"
        path = replay_dir / f"{base_name}.crdemo.gz"
        counter = 1
        while path.exists():
            path = replay_dir / f"{base_name}_{counter}.crdemo.gz"
            counter += 1
        path.write_bytes(data)
        checkpoints_path = default_checkpoints_path(path)
        dump_checkpoints_file(
            checkpoints_path,
            ReplayCheckpoints(
                version=CHECKPOINTS_FORMAT_VERSION,
                replay_sha256=digest,
                sample_rate=int(self._replay_checkpoints_sample_rate or 0),
                checkpoints=list(self._replay_checkpoints),
            ),
        )
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        if self._console is not None:
            self._console.log.log(f"replay: saved {path}")
            self._console.log.log(f"replay: saved {checkpoints_path}")
            self._console.log.flush()

    def update(self, dt: float) -> None:
        self._update_audio(dt)

        dt_frame = self._tick_frame(dt)[0]
        self._handle_input()
        if self._action == "open_pause_menu":
            return

        if self._game_over_active:
            self._update_game_over_ui(dt)
            return

        any_alive = any(player.health > 0.0 for player in self._world.players)
        sim_active = (not self._paused) and any_alive

        if not sim_active:
            self._sim_clock.reset()
            if not any_alive:
                self._enter_game_over()
            return

        ticks_to_run = self._sim_clock.advance(dt_frame)
        if ticks_to_run <= 0:
            return

        dt_tick = float(self._sim_clock.dt_tick)
        input_frame = self._build_local_inputs(dt_frame=dt_frame)
        session = self._sim_session
        if session is None:
            return
        if self._world.audio_router is not None:
            self._world.audio_router.audio = self._world.audio
            self._world.audio_router.audio_rng = self._world.audio_rng
            self._world.audio_router.demo_mode_active = self._world.demo_mode_active
        if self._world.ground is not None:
            self._world._sync_ground_settings()
            self._world.ground.process_pending()
        detail_preset = 5
        fx_toggle = 0
        if self._config is not None:
            detail_preset = int(self._config.data.get("detail_preset", 5) or 5)
            fx_toggle = int(self._config.data.get("fx_toggle", 0) or 0)
        session.detail_preset = int(detail_preset)
        session.fx_toggle = int(fx_toggle)

        for tick_offset in range(int(ticks_to_run)):
            inputs = input_frame if tick_offset == 0 else self._clear_local_input_edges(input_frame)
            recorder = self._replay_recorder
            if recorder is not None:
                tick_index = recorder.record_tick(inputs)
            else:
                tick_index = None
            tick = session.step_tick(
                dt_frame=dt_tick,
                inputs=inputs,
            )
            self._world.apply_step_result(
                tick.step,
                game_tune_started=bool(session.game_tune_started),
                apply_audio=True,
                update_camera=True,
            )
            self._rush.elapsed_ms = float(session.elapsed_ms)
            self._rush.spawn_cooldown_ms = float(session.spawn_cooldown_ms)
            world_events = tick.step.events

            if tick_index is not None:
                self._record_replay_checkpoint(
                    int(tick_index),
                    rng_marks=tick.rng_marks,
                    deaths=world_events.deaths,
                    events=world_events,
                    command_hash=str(tick.step.command_hash),
                )

            if not any(player.health > 0.0 for player in self._world.players):
                self._enter_game_over()
                break

    def _draw_game_cursor(self) -> None:
        mouse_pos = self._ui_mouse
        cursor_tex = self._ui_assets.cursor if self._ui_assets is not None else None
        draw_menu_cursor(
            self._world.particles_texture,
            cursor_tex,
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_aim_cursor(self) -> None:
        mouse_pos = self._ui_mouse
        aim_tex = self._ui_assets.aim if self._ui_assets is not None else None
        draw_aim_cursor(self._world.particles_texture, aim_tex, pos=mouse_pos)

    def draw(self) -> None:
        self._world.draw(draw_aim_indicators=(not self._game_over_active))
        self._draw_screen_fade()

        hud_bottom = 0.0
        if (not self._game_over_active) and self._hud_assets is not None:
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                self._hud_assets,
                state=self._hud_state,
                player=self._player,
                players=self._world.players,
                bonus_hud=self._state.bonus_hud,
                elapsed_ms=self._rush.elapsed_ms,
                font=self._small,
                frame_dt_ms=self._last_dt_ms,
                show_health=hud_flags.show_health,
                show_weapon=hud_flags.show_weapon,
                show_xp=hud_flags.show_xp,
                show_time=hud_flags.show_time,
                show_quest_hud=hud_flags.show_quest_hud,
                small_indicators=self._hud_small_indicators(),
            )

        if not self._game_over_active:
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            line = float(self._ui_line_height())
            self._draw_ui_text(f"rush: t={self._rush.elapsed_ms / 1000.0:6.1f}s", Vec2(x, y), UI_TEXT_COLOR)
            self._draw_ui_text(f"kills={self._creatures.kill_count}", Vec2(x, y + line), UI_HINT_COLOR)
            if self._paused:
                self._draw_ui_text("paused (TAB)", Vec2(x, y + line * 2.0), UI_HINT_COLOR)
            if self._player.health <= 0.0:
                self._draw_ui_text("game over", Vec2(x, y + line * 2.0), UI_ERROR_COLOR)

        warn_y = float(rl.get_screen_height()) - 28.0
        if self._world.missing_assets:
            warn = "Missing world assets: " + ", ".join(self._world.missing_assets)
            self._draw_ui_text(warn, Vec2(24.0, warn_y), UI_ERROR_COLOR, scale=0.8)
            warn_y -= float(self._ui_line_height(scale=0.8)) + 2.0
        if self._hud_missing:
            warn = "Missing HUD assets: " + ", ".join(self._hud_missing)
            self._draw_ui_text(warn, Vec2(24.0, warn_y), UI_ERROR_COLOR, scale=0.8)

        if not self._game_over_active:
            self._draw_aim_cursor()
        else:
            self._draw_game_cursor()
            if self._game_over_record is not None:
                self._game_over_ui.draw(
                    record=self._game_over_record,
                    banner_kind=self._game_over_banner,
                    hud_assets=self._hud_assets,
                    mouse=self._ui_mouse_pos(),
                )



from __future__ import annotations

"""Fire Bullets feature hooks for deterministic presentation output."""

from typing import Callable

from grim.geom import Vec2

from ...effects import FxQueue
from ...projectiles import ProjectileHit


def queue_large_hit_decal_streak(
    *,
    hit: ProjectileHit,
    base_angle: float,
    fx_queue: FxQueue,
    rand: Callable[[], int],
) -> None:
    """Queue the large decal streak used by Fire Bullets impact hits."""
    direction = Vec2.from_angle(base_angle)
    for _ in range(6):
        dist = float(int(rand()) % 100) * 0.1
        if dist > 4.0:
            dist = float(int(rand()) % 0x5A + 10) * 0.1
        if dist > 7.0:
            dist = float(int(rand()) % 0x50 + 0x14) * 0.1
        # Native `projectile_update` consumes one unconditional draw per loop
        # before the freeze branch (`crt_rand` @ 0x0042184c).
        rand()
        fx_queue.add_random(
            pos=hit.target + direction * (dist * 20.0),
            rand=rand,
        )
